/*
 * FunctionEvaluationTask.java
 *
 * Created on January 30, 2006, 11:52 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package gametheory;

import java.util.concurrent.*;

/**
 * Parallelizable Task that evaluates the moments for a given Game.
 *
 * @author Stephen P. Ryan
 */
public class FunctionEvaluationTaskGMM implements Callable<Jama.Matrix> {

    /**
     * Number of moments.
     */
    private int numMoments;
    /**
     * Game to be evaluated.
     */
    private Game g;
    /**
     * Vector of unknown parameters.
     */
    private double[] x;
    /**
     * Switch to impose upper and lower bounds on the calculated probabilities.
     */
    private static boolean censoring = true;

    /**
     * Constructor for GMM Task.
     *
     * @param numMoments Number of moments.
     * @param g Game to evaluate.
     * @param x Vector of unknown parameters.
     */
    public FunctionEvaluationTaskGMM(int numMoments, Game g, double[] x) {
        this.g = g;
        this.numMoments = numMoments;
        this.x = x;
    }

    /**
     * Computational logic.
     *
     * @return Fitted moment for this Game.
     */
    public Jama.Matrix call() {
        Jama.Matrix outcome;
        Jama.Matrix action;
        double norm;
        double probG;
        double indicator;
        Jama.Matrix moments = new Jama.Matrix(numMoments, 1);

        /**
         * Loop over moments.
         */
        for (int j = 0; j < numMoments; j++) {
            indicator = 0;
            outcome = g.getOutcome();
            /**
             * Obtain action associated with this moment.
             */
            action = g.getActionList().get(j);
            /**
             * Determine if this outcome was played in the data.
             */
            norm = (outcome.minus(action)).norm2();
            if (norm == 0) {
                indicator = 1;
            }
            /**
             * Compute probability of observing outcome.
             */
            probG = g.getProbabilityAction(j, x);
            /**
             * Impose bounds on probabilities if desired.
             */
            if (censoring) {
                probG = Math.min(1, probG);
                probG = Math.max(0, probG);
            }
            /**
             * Compose moment vector if probability is not infinite or NaN
             */
            if (!Double.isNaN(probG) && !Double.isInfinite(probG)) {
                moments.set(j, 0, moments.get(j, 0) + Math.pow(indicator - probG, 2));
            }
        }
        /**
         * Return value.
         */
        return moments;
    }
}
