/*
 * Game.java
 *
 * Created on December 26, 2005, 4:50 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package gametheory;

import JSci.maths.statistics.NormalDistribution;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.Serializable;
import java.util.ArrayList;
import utility.pmUtility;

/**
 * Contains all the information for a discrete game of complete information: actions, payoffs, importance games, and equilibria.
 *
 * @author Stephen P. Ryan
 */
public class Game implements Serializable {

    /**
     * List of action permutations.
     */
    private ArrayList<Jama.Matrix> actionList;
    /**
     * List of deterministic payoffs associated with each action in actionList.
     */
    private ArrayList<Jama.Matrix> payoffList;
    /**
     * Number of players.
     */
    private int numPlayers;
    /**
     * Number of actions.
     */
    private int numActions;
    /**
     * Standard normal distribution.
     */
    static NormalDistribution normal = new NormalDistribution(0, 1);
    /**
     * Set of equilibria associated with this game.
     */
    private ArrayList<double[]> equilibria;
    /**
     * Simulated outcome.
     */
    private Jama.Matrix outcome;
    /**
     * List of importance games associated with this game.
     */
    private ArrayList<Game> importanceGamesList;
    /**
     * Number of mixed strategy equilibria.
     */
    private int numMixedStrategyEquilibria;
    /**
     * Number of pure strategy equilibria.
     */
    private int numberPureStrategyEquilibria;
    /**
     * Array enumerating if an equilibrium is in pure strategies or not.
     */
    private boolean[] pureEnumeration;
    /**
     * Index of the most efficient equilibrium.
     */
    private int efficientIndex;
    /**
     * Array enumerating if an equilibrium is dominated or not.
     */
    private boolean[] dominatedEnumeration;
    /**
     * ArrayList containing utilities.
     */
    private ArrayList<Jama.Matrix> utilityList;

    /** Creates a new instance of Game
     * @param numPlayers Number of players.
     * @param numActions Number of actions.
     */
    public Game(int numPlayers, int numActions) {
        this.actionList = GameTheoryConstants.actionList;
        this.numPlayers = numPlayers;
        this.numActions = numActions;
        payoffList = new ArrayList<Jama.Matrix>();
        generateEmptyPayoffList();
    }

    /**
     * Creates a deep copy of the current game.
     *
     * @return A deep copy clone of the current game.
     */
    public Game deepCopy() {
        Game g = new Game(numPlayers, numActions);
        ArrayList<Jama.Matrix> payoffList2 = new ArrayList<Jama.Matrix>();
        ArrayList<double[]> equList2 = new ArrayList<double[]>();
        ArrayList<Jama.Matrix> euList = new ArrayList<Jama.Matrix>();
        boolean[] pure = null;
        if (pureEnumeration != null) {  // some games have not had their equilibria solved
            pure = new boolean[pureEnumeration.length];
            for (int i = 0; i < equilibria.size(); i++) {
                double[] prob = equilibria.get(i);
                equList2.add(prob);
                pure[i] = pureEnumeration[i];
            }
        }
        boolean[] dominated = null;
        if (dominatedEnumeration != null) {
            dominated = new boolean[dominatedEnumeration.length];
            for (int i = 0; i < equilibria.size(); i++) {
                dominated[i] = dominatedEnumeration[i];
            }
        }
        for (int i = 0; i < payoffList.size(); i++) {
            Jama.Matrix payoff = (payoffList.get(i)).copy();
            payoffList2.add(payoff);
            if (utilityList != null) {
                Jama.Matrix utility = (utilityList.get(i)).copy();
                euList.add(utility);
            }
        }
        g.setEfficientIndex(efficientIndex);
        g.setImportanceGamesList(importanceGamesList);
        g.setPayoffList(payoffList2);
        g.setUtilityList(euList);
        g.setNumMixedEqu(numMixedStrategyEquilibria);
        g.setNumPureEqu(numberPureStrategyEquilibria);
        g.setEquilibria(equList2);
        g.setDominatedEnumeration(dominated);
        return g;
    }

    /**
     * Computes the expected value of an equilibrium; used for finding the efficient equilibrium in the set.
     *
     * @param equilibriumNumber Index of the equilibrium the expected value is to be calculated for.
     * @return Expected value of the equilibrium.
     */
    public double getExpectedValueEquilibrium(int equilibriumNumber) {
        double value = 0;
        double[] prob = equilibria.get(equilibriumNumber);

        for (int i = 0; i < actionList.size(); i++) {
            Jama.Matrix action = actionList.get(i);
            Jama.Matrix utility = utilityList.get(i);
            double innerProb = 1;
            double innerPayoffSum = 0;
            for (int j = 0; j < action.getRowDimension(); j++) {
                if (action.get(j, 0) == 0) {
                    innerProb *= prob[2 * j];
                } else {
                    innerProb *= 1.0 - prob[2 * j];
                }
                innerPayoffSum += utility.get(j, 0);
            }
            value += innerPayoffSum * innerProb;
        }
        return value;
    }

    /**
     * Calculate the expected value of a particular equilibrium for a specific player.
     *
     * @param equilibriumNumber Index of equilibrium to evaluate.
     * @param playerIndex Index of player to evaluate.
     * @return Expected value of that equilibrium for a given player.
     */
    public double getExpectedValueEquilibrium(int equilibriumNumber, int playerIndex) {
        double value = 0;
        double[] prob = equilibria.get(equilibriumNumber);

        for (int i = 0; i < actionList.size(); i++) {
            Jama.Matrix action = actionList.get(i);
            Jama.Matrix utility = utilityList.get(i);

            double innerProb = 1;
            for (int j = 0; j < action.getRowDimension(); j++) {
                if (action.get(j, 0) == 0) {
                    innerProb *= prob[2 * j];
                } else {
                    innerProb *= 1.0 - prob[2 * j];
                }
            }
            double innerPayoffSum = utility.get(playerIndex, 0);
            value += innerPayoffSum * innerProb;
        }
        return value;
    }

    /**
     * Compute the efficient equilibrium and store the result through <b>setEfficientIndex</b>.
     */
    public void findEfficientEquilibrium() {
        double ev = 0;
        setEfficientIndex(0);
        for (int i = 0; i < equilibria.size(); i++) {
            double evNew = getExpectedValueEquilibrium(i);
            if (evNew > ev || i == 0) {
                ev = evNew;
                setEfficientIndex(i);
            }
        }
    }

    /**
     * Simulate an outcome from the game.
     *
     * @return Matrix containing the equilibrium actions of all players.
     */
    public Jama.Matrix getSimulatedOutcome() {
        /**
         * Generate a probability distribution over equilibria.
         */
        double[] probEqu = new double[equilibria.size()];
        double sumEqu = 0;
        for (int i = 0; i < equilibria.size(); i++) {
            double inner = 0;
            if (pureEnumeration[i]) {
                inner += GameTheoryConstants.PURE_STRATEGY_EQUILIBRIUM_SHIFTER;
            }
            if (i == efficientIndex) {
                inner += GameTheoryConstants.EFFICIENT_EQUILIBRIUM_SHIFTER;
            }
            if (dominatedEnumeration[i]) {
                inner += GameTheoryConstants.DOMINATED_EQUILIBRIUM_SHIFTER;
            }
            probEqu[i] = Math.exp(inner);
            sumEqu += Math.exp(inner);
        }
        for (int i = 0; i < equilibria.size(); i++) {
            probEqu[i] /= sumEqu;
        }
        /**
         * Compute which equilibrium is played.
         */
        double draw = Math.random();
        double sumProb = 0;
        int equilibriumSelected = 0;
        while (draw > sumProb) {
            if (draw > sumProb) {
                sumProb += probEqu[equilibriumSelected];
                if (draw > sumProb) {
                    equilibriumSelected++;
                }
            }
        }
        /**
         * Now we know which equilibrium we are playing, draw an outcome from that equilibrium
         */
        Jama.Matrix simulatedOutcome = new Jama.Matrix(numPlayers, 1);
        double[] equProb = (double[]) equilibria.get(equilibriumSelected);
        for (int i = 0; i < numPlayers; i++) {
            draw = Math.random();
            if (draw > equProb[i * 2]) {
                simulatedOutcome.set(i, 0, 1);
            }
        }
        outcome = simulatedOutcome.copy();
        return simulatedOutcome;
    }

    /**
     * Compute a single moment for this game.  Equivalently, this method returns the probability
     * of observing a given outcome.  The method computes this probability by summing over importance
     * games and computing the probability of observing the outcome under each of the importance
     * game's equilibria.
     *
     * @param actionIndex Index of action to compute the moment for.
     * @param x Vector of unknown parameters.
     * @return Probability of seeing that action played in the data.
     */
    public double getProbabilityAction(int actionIndex, double[] x) {
        double prob = 0;
        /**
         * Actual outcome observed in the data.
         */
        Jama.Matrix action = actionList.get(actionIndex);
        /**
         * Sum over importance games to compute the probability of seeing a particular outcome.
         */
        for (int s = 0; s < importanceGamesList.size(); s++) {
            /**
             * Obtain the importance game.
             */
            Game importanceGame = importanceGamesList.get(s);
            double innerProb = 0;
            /**
             * Obtain the set of equilibria for this importance game.
             */
            ArrayList<double[]> importanceGameEquilibria = importanceGame.getEquilibria();
            for (int i = 0; i < importanceGameEquilibria.size(); i++) {
                /**
                 * Compute the probability of this equilibrium being played in this importance game.
                 */
                double lambda = importanceGame.getLambda(i, x);
                double probA = 1;
                double[] strategy = importanceGameEquilibria.get(i);
                /**
                 * Compute the probability of the action given this equilibrium.
                 */
                for (int j = 0; j < numPlayers; j++) {
                    if (action.get(j, 0) == 0) {
                        probA *= strategy[j * 2];
                    } else {
                        probA *= strategy[j * 2 + 1];
                    }
                }
                /**
                 * Probability of observing this action is the product of the probability
                 * of the equilibrium being played and the probability of the outcome
                 * under that equilibrium.
                 */
                innerProb += lambda * probA;
            }
            /**
             * Need to weight this probability appropriately by the importance weight.
             */
            double ratio = importanceGame.getHQRatio(x, payoffList);
            innerProb *= ratio;
            /**
             * Add to total probability.
             */
            prob += innerProb;
        }
        /**
         * Normalize by number of importance games.
         */
        prob *= 1.0 / importanceGamesList.size();
        /**
         * Return result.
         */
        return prob;
    }

    /**
     * Computes the probability of an equilibrium being selected in a play of the game.
     *
     * @param equilibriumIndex Index of equilibrium to be evaluated.
     * @param x Vector of unknown parameters.
     * @return Probability that equilibrium will be selected in a play of the game.
     */
    public double getLambda(int equilibriumIndex, double[] x) {
        /**
         * Define array containing equilibria probabilities.
         */
        double[] probEqu = new double[equilibria.size()];
        double sumEqu = 0;
        /**
         * Compute numerators for each equilibrium.
         */
        for (int i = 0; i < equilibria.size(); i++) {
            double inner = 0;
            if (pureEnumeration[i]) {
                inner += x[2];
            }
            if (efficientIndex == i) {
                inner += x[3];
            }
            if (dominatedEnumeration[i]) {
                // System.out.println("adding dominated shifter");
                inner += x[4];
            } else {
                // System.out.println("adding not dominated");
            }
            probEqu[i] = Math.exp(inner);
            sumEqu += Math.exp(inner);
        }
        /**
         * Divide all probabilities by the denominator.
         */
        for (int i = 0; i < equilibria.size(); i++) {
            probEqu[i] /= sumEqu;
        }
        /**
         * Return the result.
         */
        return probEqu[equilibriumIndex];
    }

    /**
     * Computes the importance weight for the current game.
     *
     * @param x Vector of unknown parameters.
     * @param expectedProfitsList ArrayList of deterministic expected profits.
     * @return Importance weight for this game.
     */
    public double getHQRatio(double[] x, ArrayList<Jama.Matrix> expectedProfitsList) {
        double prob = 1;
        /**
         * Indexed over the set of all possible actions (number of boxes in the normal form game)
         */
        for (int i = 0; i < actionList.size(); i++) {
            Jama.Matrix expectedProfits = expectedProfitsList.get(i);
            Jama.Matrix utility = utilityList.get(i);
            Jama.Matrix action = actionList.get(i);
            /**
             * Within each box, index over all players (equal to row dimension of expectedProfits)
             */
            for (int j = 0; j < expectedProfits.getRowDimension(); j++) {
                /**
                 * Compute the difference in the model prediction against the importance density utility,
                 * which is based on a fixed initial guess of the underlying parameters.
                 *
                 * First, compute utility of profits
                 */
                double f = x[5] * expectedProfits.get(j, 0);
                /**
                 * Do the same for the importance game.  Note that the constant referred to here is where
                 * the initial guess for the parameters in the importance game comes in.
                 */
                double qf = GameTheoryConstants.PROFIT_MULTIPLIER * expectedProfits.get(j, 0);
                /**
                 * Shift utilities by bid costs.
                 */
                if (action.get(j, 0) == 1) {
                    if (j == 0) {
                        f -= x[1];
                        qf -= GameTheoryConstants.BID_COST1;
                    }
                    if (j == 1) {
                        f -= x[6];
                        qf -= GameTheoryConstants.BID_COST2;
                    }
                    if (j == 2) {
                        f -= x[7];
                        qf -= GameTheoryConstants.BID_COST3;
                    }
                    if (j == 3) {
                        f -= x[8];
                        qf -= GameTheoryConstants.BID_COST4;
                    }
                }
                /**
                 * Multiply and divide by the current probability and the fixed probability to obtain the importance
                 * weight.
                 */
                double innerProb = 1;
                innerProb *= normal.probability(f - utility.get(j, 0));
                innerProb /= normal.probability(qf - utility.get(j, 0));
                prob *= innerProb;
            }
        }
        return prob;
    }

    /**
     * Computes utilities for use in importance games and Monte Carlo.
     */
    public void calculateUtilities() {
        utilityList = new ArrayList<Jama.Matrix>();
        for (int i = 0; i < actionList.size(); i++) {
            Jama.Matrix payoff = payoffList.get(i);
            Jama.Matrix action = actionList.get(i);
            Jama.Matrix utility = new Jama.Matrix(payoff.getRowDimension(), payoff.getColumnDimension(), 0);

            for (int j = 0; j < payoff.getRowDimension(); j++) {
                double bidCost = 0;
                if (action.get(j, 0) == 1) {
                    if (j == 0) {
                        bidCost = GameTheoryConstants.BID_COST1;
                    }
                    if (j == 1) {
                        bidCost = GameTheoryConstants.BID_COST2;
                    }
                    if (j == 2) {
                        bidCost = GameTheoryConstants.BID_COST3;
                    }
                    if (j == 3) {
                        bidCost = GameTheoryConstants.BID_COST4;
                    }
                }
                utility.set(j, 0, GameTheoryConstants.PROFIT_MULTIPLIER * payoff.get(j, 0) - bidCost + normal.inverse(Math.random()));
            }
            utilityList.add(utility);
        }
    }

    /**
     * Creates a new, empty payoff list.
     */
    public void generateEmptyPayoffList() {
        for (int i = 0; i < actionList.size(); i++) {
            Jama.Matrix payoffVector = actionList.get(i);
            payoffList.add(payoffVector.times(1.0));
        }
    }

    /**
     * Calculates the index of an action in the actionList.
     *
     * @param findMe Jama.Matrix representation of the action to find.
     * @return The index of the action in actionList.
     */
    public int findIndexAction(Jama.Matrix findMe) {
        // this will return the index of the specified action by doing the search myself
        int index = -1;
        for (int i = 0; i < actionList.size(); i++) {
            Jama.Matrix action = actionList.get(i);
            boolean same = true;
            for (int j = 0; j < action.getRowDimension(); j++) {
                if (action.get(j, 0) != findMe.get(j, 0)) {
                    same = false;
                }
            }
            if (same) {
                index = i;
                i = actionList.size();
            }
        }
        return index;
    }

    /**
     * Output list of all actions to the console.
     */
    public void outputActions() {
        for (int i = 0; i < actionList.size(); i++) {
            pmUtility.prettyPrintVector( actionList.get(i));
        }
    }

    /**
     * Sets payoff for specific player and action.
     *
     * @param action Action to assign payoff to.
     * @param playerIndex Index of player.
     * @param payoff Value to assign.
     * @param report Switch to print debugging information.
     */
    public void setPayoff(Jama.Matrix action, int playerIndex, double payoff, boolean report) {
        /**
         * Payoffs are the expected engineering profits conditional on winning the job, without bid costs
         */
        int index = findIndexAction(action);
        if (report) {
            pmUtility.prettyPrintVector(action);
            System.out.println(index);
        }
        Jama.Matrix payoffVector = payoffList.get(index);
        payoffVector.set(playerIndex, 0, payoff);
    }

    /**
     * Returns string form of utilityList.
     *
     * @return String form of utilityList.
     */
    public String getUtilitiesFlatString() {
        String flatList = "";

        for (int i = 0; i < payoffList.size(); i++) {
            Jama.Matrix utility = utilityList.get(i);
            flatList = flatList.concat(stringVector(utility));
        }
        return flatList;
    }

    /**
     * Returns a string representation of the game.
     *
     * @return String representation of the game.
     */
    @Override
    public String toString() {
        String result = "NFG 1 R \"BHR Game\"\n";
        result = result.concat("{ ");
        for (int i = 0; i < numPlayers; i++) {
            result = result.concat("\"Player " + i + "\" ");
        }
        result = result.concat("} { ");
        for (int i = 0; i < numPlayers; i++) {
            result = result.concat(numActions + " ");
        }
        result = result.concat("}\n");
        result = result.concat(getUtilitiesFlatString() + "\n");
        return result;
    }

    /**
     * Convert Matrix to string.
     *
     * @param x Matrix to convert.
     * @return String form of Matrix.
     */
    public String stringVector(Jama.Matrix x) {
        String result = "";
        for (int i = 0; i < x.getRowDimension(); i++) {
            result = result.concat((x.get(i, 0)) + " ");
        }
        return result;
    }

    /**
     * Print game to filename.
     *
     * @param filename Name of file to output game to.
     */
    public void outputGame(String filename) {
        try {
            BufferedWriter out = new BufferedWriter(new FileWriter(filename));
            out.write(toString());
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Get list of actions.
     *
     * @return List of actions.
     */
    public ArrayList<Jama.Matrix> getActionList() {
        return actionList;
    }

    /**
     * Get deterministic payoff list.
     *
     * @return Deterministic payoff list.
     */
    public ArrayList<Jama.Matrix> getPayoffList() {
        return payoffList;
    }

    /**
     * Get number of players.
     *
     * @return Number of players.
     */
    public int getNumPlayers() {
        return numPlayers;
    }

    /**
     * Get the number of actions.
     *
     * @return Number of actions.
     */
    public int getNumActions() {
        return numActions;
    }

    /**
     * Obtain the set of equilibria associated with this game.
     *
     * @return The set of equilibria associated with this game.
     */
    public ArrayList<double[]> getEquilibria() {
        return equilibria;
    }

    /**
     * Set equilibria to this game and enumerate their types (dominated, efficient, mixed/pure)
     *
     * @param equilibria Array containing equilibria to be enumerated.
     */
    public void setEquilibria(ArrayList<double[]> equilibria) {
        this.equilibria = equilibria;
        // this is where i should compute all enumerations of equilibria (num pure, num mixed, efficient, etc.)
        numMixedStrategyEquilibria = 0;
        numberPureStrategyEquilibria = 0;
        /**
         * Enumerate pure/mixed.
         */
        pureEnumeration = new boolean[equilibria.size()];
        for (int i = 0; i < equilibria.size(); i++) {
            double[] equ = (double[]) equilibria.get(i);
            pureEnumeration[i] = true;
            for (int j = 0; j < equ.length; j++) {
                if (equ[j] != 0 && equ[j] != 1) {
                    pureEnumeration[i] = false;
                }
            }
            if (pureEnumeration[i]) {
                numberPureStrategyEquilibria++;
            } else {
                numMixedStrategyEquilibria++;
            }
        }
        /**
         * Enumerate dominated.
         */
        dominatedEnumeration = new boolean[equilibria.size()];
        for (int i = 0; i < equilibria.size(); i++) {
            double[] epayoff = new double[GameTheoryConstants.NUM_PLAYERS];
            for (int k = 0; k < GameTheoryConstants.NUM_PLAYERS; k++) {
                epayoff[k] = getExpectedValueEquilibrium(i, k);
            }
            dominatedEnumeration[i] = false;
            for (int j = 0; j < equilibria.size(); j++) {
                boolean dominated = true;
                double[] eother = new double[GameTheoryConstants.NUM_PLAYERS];
                for (int k = 0; k < GameTheoryConstants.NUM_PLAYERS; k++) {
                    eother[k] = getExpectedValueEquilibrium(j, k);
                }
                for (int k = 0; k < GameTheoryConstants.NUM_PLAYERS; k++) {
                    if (epayoff[k] > eother[k]) {
                        dominated = false;
                    }
                }
                if (dominated && i != j) {
                    dominatedEnumeration[i] = true;
                    j = equilibria.size();
                }
            }
        }
        /**
         * Find the efficient equilibrium.
         */
        findEfficientEquilibrium();
    }

    /**
     * Get the simulated outcome.
     *
     * @return The simulated outcome.
     */
    public Jama.Matrix getOutcome() {
        return outcome;
    }

    /**
     * Set simulated outcome.
     *
     * @param outcome The simulated outcome.
     */
    public void setOutcome(Jama.Matrix outcome) {
        this.outcome = outcome;
    }

    /**
     * Get the importance games list.
     *
     * @return The list of importance games.
     */
    public ArrayList<Game> getImportanceGamesList() {
        return importanceGamesList;
    }

    /**
     * Set importance games list.
     *
     * @param aImportanceGamesList The list of importance games.
     */
    public void setImportanceGamesList(ArrayList<Game> aImportanceGamesList) {
        importanceGamesList = aImportanceGamesList;
    }

    /**
     * Get the number of mixed strategy equilibria.
     *
     * @return The number of mixed strategy equilibria.
     */
    public int getNumMixedEqu() {
        return numMixedStrategyEquilibria;
    }

    /**
     * Set the number of mixed strategy equilibria.
     *
     * @param numMixedEqu Number of mixed strategy equilibria.
     */
    public void setNumMixedEqu(int numMixedEqu) {
        this.numMixedStrategyEquilibria = numMixedEqu;
    }

    /**
     * Get the number of pure strategy equilibria.
     *
     * @return The number of pure strategy equilibria.
     */
    public int getNumPureEqu() {
        return numberPureStrategyEquilibria;
    }

    /**
     * Set the number of pure strategy equilibria.
     *
     * @param numPureEqu Number of pure strategy equilibria.
     */
    public void setNumPureEqu(int numPureEqu) {
        this.numberPureStrategyEquilibria = numPureEqu;
    }

    /**
     * Set the payoff List.
     *
     * @param payoffList ArrayList<Jama.Matrix> of payoffs.
     */
    public void setPayoffList(ArrayList<Jama.Matrix> payoffList) {
        this.payoffList = payoffList;
    }

    /**
     * Get enumeration of pure strategies.
     *
     * @return Enumeration of pure strategies.
     */
    public boolean[] getPureEnumeration() {
        return pureEnumeration;
    }

    /**
     * Set enumeration of pure strategies.
     *
     * @param pureEnumeration Enumeration of pure strategies.
     */
    public void setPureEnumeration(boolean[] pureEnumeration) {
        this.pureEnumeration = pureEnumeration;
    }

    /**
     * Get index of the efficient equilibrium.
     *
     * @return Index of the efficient equilibrium.
     */
    public int getEfficientIndex() {
        return efficientIndex;
    }

    /**
     * Set index of the efficient equilibrium.
     *
     * @param efficientIndex Index of the efficient equilibrium.
     */
    public void setEfficientIndex(int efficientIndex) {
        this.efficientIndex = efficientIndex;
    }

    /**
     * Get the utility list.
     *
     * @return The list of utilities.
     */
    public ArrayList getUtilityList() {
        return utilityList;
    }

    /**
     * Set the utility list.
     *
     * @param utilityList The list of utilities.
     */
    public void setUtilityList(ArrayList<Jama.Matrix> utilityList) {
        this.utilityList = utilityList;
    }

    /**
     * Get enumeration of dominated strategies.
     *
     * @return The enumeration of dominated strategies.
     */
    public boolean[] getDominatedEnumeration() {
        return dominatedEnumeration;
    }

    /**
     * Set enumeration of dominated strategies.
     *
     * @param dominatedEnumeration Enumeration of dominated strategies.
     */
    public void setDominatedEnumeration(boolean[] dominatedEnumeration) {
        this.dominatedEnumeration = dominatedEnumeration;
    }
}
