/*
 * GameMinimization.java
 *
 * Created on January 17, 2006, 11:24 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package gametheory;

import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import mcmc.*;
import optimization.Uncmin_methods;
import utility.pmUtility;

/**
 * Class to perform the optimization of the GMM function on real auction entry data.
 *
 * @author Stephen P. Ryan
 */
public class GameMinimization implements mcmcFunction, Uncmin_methods {

    /**
     * List of Games (observations).
     */
    ArrayList<Game> gameList;
    /**
     * Impose limits on probabilities of outcomes.
     */
    boolean censoring = true;
    /**
     * ExecutorService to use for distributing tasks to computing cores.  This is easily replaced with
     * Gridgain's executor service if using a distributed computation grid.
     */
    ExecutorService tpes;
    /**
     * Number of threads to use in the ExecutorService.
     */
    int numWorkers = GameTheoryConstants.NUMBER_THREADS;

    /** Creates a new instance of GameMinimization
     * @param gameList List of games (observations).
     */
    public GameMinimization(ArrayList<Game> gameList) {
        this.gameList = gameList;
        tpes = Executors.newFixedThreadPool(numWorkers);
    }

    /**
     * Use Chernozhukov and Hong's (2003) LTE method to optimize the GMM function.
     *
     * @param guess Initial guess of parameters.
     * @return MCMC chain containing results of LTE procedure.
     */
    public Jama.Matrix executeLTE(double[] guess) {
        gibbsLTEGeneralized lte = new gibbsLTEGeneralized(this, 500, 500, guess);
        System.out.print("f_min: " + f_to_minimize(lte.getLowestPoint()) + "\n");
        pmUtility.prettyPrint(new Jama.Matrix(lte.getLowestPoint(), 1));
        guess = lte.getLowestPoint();
        return lte.getChain();
    }

    /**
     * Use Newton-based solver to optimize GMM function.  Newton methods all use
     * the translations of FORTRAN Uncmin methods done by Steve Verrill.  See his
     * webpage at <a href="http://www1.fpl.fs.fed.us/optimization.html>FPL</a>
     * for code and documentation.
     *
     * @param guess Initial guess of unknown parameters.
     * @return Solution from minimization problem.
     */
    public double[] execute(double[] guess) {
        int numParams = GameTheoryConstants.NUM_PARAMS;
        double[] xpls = new double[numParams + 1];
        double[] fpls = new double[2];
        double[] gpls = new double[numParams + 1];
        int[] itrmcd = new int[2];
        double[][] a = new double[numParams + 1][numParams + 1];
        double[] udiag = new double[numParams + 1];
        double[] typsiz = new double[numParams + 1];
        double[] fscale = {0, 1E-8};
        int[] method = {0, 1};  // 1 line search 2 double dogleg 3 more-hebdon
        int[] iexp = {0, 0}; // 1 if expensive to evaluate
        int[] msg = {0, 0};
        int[] ndigit = {0, 8}; // number of good digits in minimand
        int[] itnlim = {0, 150};
        int[] iagflg = {0, 0};
        int[] iahflg = {0, 0};
        double[] dlt = {0, 1};
        double[] gradtl = {0, 1E-8};
        double[] stepmx = {0, 3.0};
        double[] steptl = {0, 1E-8};

        optimization.Uncmin_f77 minimizer = new optimization.Uncmin_f77(false);
        minimizer.optif9_f77(numParams, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);

        return guess;
    }

    /**
     * MCMC objective function for use in the LTE.
     *
     * @param x Vector of unknown parameters.
     * @return Objective function in LTE form.
     */
    public double objectiveFunction(double[] x) {
        double numObs = gameList.size();
        return -numObs * f_to_minimize(x) / 2.0;
    }

    /**
     * LTE probability of proposed parameters.
     *
     * @param x Proposed parameter vector.
     * @return One everywhere, implying no <i>a priori</i> restrictions on permissable parameters
     */
    public double pi(double[] x) {
        return 1;
    }

    /**
     * Computes GMM objective function.
     *
     * @param x Vector of unknown parameters.
     * @return Value of GMM function at a guess of the parameters.
     */
    public double f_to_minimize(double x[]) {
        /**
         * Compute number of moments.
         */
        int numMoments = (gameList.get(0)).getActionList().size() - 1;
        Jama.Matrix moments = new Jama.Matrix(numMoments, 1);
        /**
         * Initialize Future list for holding results of distributed Tasks
         */
        ArrayList<Future<Jama.Matrix>> workerList = new ArrayList<Future<Jama.Matrix>>();

        try {
            /**
             * Submit tasks.
             */
            for (int i = 0; i < gameList.size(); i++) {
                workerList.add(tpes.submit(new FunctionEvaluationTaskGMM(numMoments, gameList.get(i), x)));
            }
            /**
             * Compute moments by summing over moment of each agent.
             */
            for (int i = 0; i < gameList.size(); i++) {
                moments.plusEquals(workerList.get(i).get());
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        /**
         * Divide moment by number of observations.
         */
        moments.timesEquals(1.0 / gameList.size());
        /**
         * Compute sandwich form.
         */
        double value = ((moments.transpose()).times(moments)).get(0, 0);
        /**
         * Return value.
         */
        return value;
    }

    /**
     * User-supplied gradient for Uncmin.  Not used.
     *
     * @param x Parameters.
     * @param g Gradient.
     */
    public void gradient(double x[], double g[]) {
    }

    /**
     * User-supplied Hessian for Uncmin.  Not used.
     *
     * @param x Parameters.
     * @param h Hessian.
     */
    public void hessian(double x[], double h[][]) {
    }
}
