/*
 * GameMinimizationPrivateInformation.java
 *
 * Created on January 27, 2006, 1:18 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package gametheory;

import JSci.maths.statistics.NormalDistribution;
import java.util.ArrayList;
import optimization.Uncmin_methods;
import utility.pmUtility;

/**
 * Private information version of discrete game used to find starting values
 * in the complete information game.
 *
 * @author Stephen P. Ryan
 */
public class GameMinimizationPrivateInformation implements Uncmin_methods {

    /**
     * Switch to impose bounds on probabilities.
     */
    private boolean censor = false;
    /**
     * List of Games (observations).
     */
    private ArrayList<Game> gameList;
    /**
     * Standard normal distribution.
     */
    private NormalDistribution normal = new NormalDistribution();

    /** Creates a new instance of GameMinimizationPrivateInformation
     * @param gameList List of Games (observations).
     */
    public GameMinimizationPrivateInformation(ArrayList<Game> gameList) {
        this.gameList = gameList;
    }

    /**
     * Compute solution to private information game using Newton-based methods.
     *
     * @return Best solution vector.
     */
    public double[] execute() {
        int numParams = 2;  // bid cost (1), scaling factor on profits (2)
        double[] guess = new double[numParams + 1];
        double[] xpls = new double[numParams + 1];
        double[] fpls = new double[2];
        double[] gpls = new double[numParams + 1];
        int[] itrmcd = new int[2];
        double[][] a = new double[numParams + 1][numParams + 1];
        double[] udiag = new double[numParams + 1];
        double[] typsiz = new double[numParams + 1];
        double[] fscale = {0, 1E-15};
        int[] method = {0, 1};  // 1 line search 2 double dogleg 3 more-hebdon
        int[] iexp = {0, 0}; // 1 if expensive to evaluate
        int[] msg = {0, 0};
        int[] ndigit = {0, 8};
        int[] itnlim = {0, 150};
        int[] iagflg = {0, 0};
        int[] iahflg = {0, 0};
        double[] dlt = {0, 1};
        double[] gradtl = {0, 1E-15};
        double[] stepmx = {0, 1E8};
        double[] steptl = {0, 1E-15};

        /**
         * Start with zero vector.
         */
        for (int i = 0; i < numParams; i++) {
            guess[i + 1] = 0;
        }

        optimization.Uncmin_f77 minimizer = new optimization.Uncmin_f77(true);
        minimizer.optif9_f77(numParams, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);

        return guess;
    }

    /**
     * Objective function in private information case.
     *
     * @param x Vector of parameters.
     * @return Value of objective function.
     */
    public double f_to_minimize(double x[]) {
        double value = 0;
        double count = 0;
        for (int i = 0; i < gameList.size(); i++) {
            Game g = gameList.get(i);
            // go through for each game, compute likelihood of observed action vector given covariates and guess for costs
            ArrayList<Jama.Matrix> actionList = g.getActionList();
            ArrayList<Jama.Matrix> payoffList = g.getPayoffList();
            Jama.Matrix outcome = g.getOutcome();
            for (int p = 0; p < g.getNumPlayers(); p++) {
                // for each player, compute expected value of entering (not entering profits are 0, which is convienent)
                double expectedValuePlayer = 0;
                for (int j = 0; j < actionList.size(); j++) {
                    Jama.Matrix action = actionList.get(j);
                    Jama.Matrix payoff = payoffList.get(j);
                    double probAction = 1;
                    if (action.get(p, 0) == 1) {
                        for (int pother = 0; pother < g.getNumPlayers(); pother++) {
                            if (pother != p) {
                                double prob = normal.cumulative(x[2] * payoff.get(pother, 0) - x[1]);
                                if (censor) {
                                    prob = Math.max(1E-16, prob);
                                    prob = Math.min(1.0 - 1E-16, prob);
                                }
                                if (Double.isNaN(prob)) {
                                    System.out.println("NaN problem: " + pmUtility.stringPrettyPrint(new Jama.Matrix(x, 1)));
                                }
                                if (Double.isInfinite(prob)) {
                                    System.out.println("Infinity problem: " + pmUtility.stringPrettyPrint(new Jama.Matrix(x, 1)));
                                }

                                if (action.get(pother, 0) == 0) {
                                    probAction *= (1.0 - prob);
                                } else {
                                    probAction *= prob;
                                }
                            }
                        }
                    }
                    double profit = x[2] * payoff.get(p, 0) - x[1];
                    expectedValuePlayer += probAction * profit;
                }
                double probEntry = normal.cumulative(expectedValuePlayer);
                if (outcome.get(p, 0) == 0) {
                    value += Math.log(1.0 - probEntry);
                } else {
                    value += Math.log(probEntry);
                }
                count++;
            }
        }
        value /= count;
        return -value;
    }

    /**
     * User-supplied gradient, not used.
     *
     * @param x Parameters.
     * @param g Gradient.
     */
    public void gradient(double x[], double g[]) {
    }

    /**
     * User-supplied Hessian, not used.
     *
     * @param x Parameters.
     * @param h Hessian.
     */
    public void hessian(double x[], double h[][]) {
    }
}
