/*
 * GameTheoryGUI.java
 *
 * Created on January 5, 2006, 11:23 AM
 */
package gametheory;

import JSci.maths.statistics.NormalDistribution;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.ObjectOutputStream;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.UIManager;
import javax.swing.table.DefaultTableModel;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import utility.pmUtility;

/**
 * GUI frontend and control logic for the estimation procedure in Bajari, Hong, and Ryan (2009).  The GUI consists of several
 * elements to load in the data, obtain starting values from the private information analogue of the entry game,
 * generate importance games, and run the actual estimation.
 *
 * Use of Netbeans' GUI builder is highly recommended for understand what GUI elements link to which functionality.
 *
 * @author  Stephen P. Ryan
 */
public class GameTheoryGUI extends javax.swing.JFrame {

    /**
     * ArrayList containing the data.
     */
    ArrayList<Game> gameList;
    /**
     * ArrayList containing the importance games used in the estimation.
     */
    ArrayList<Game> importanceGamesList;
    /**
     * ArrayList containing Monte Carlo data.  Note that there is a separate program for generating the Monte Carlos as found in the paper.
     */
    ArrayList<Game> monteCarloData;
    /**
     * Standard normal distribution.
     */
    NormalDistribution normal = new NormalDistribution();
    /**
     * ExecutorService used to dispatch worker threads to multiple CPU's.  Unlike the Monte Carlo, this version is not set up to run on a Grid, although it is a simple extension.
     * See the source code of the Monte Carlo program for details.
     */
    ExecutorService tpes = Executors.newFixedThreadPool(GameTheoryConstants.NUMBER_THREADS);

    /** Creates new form GameTheoryGUI */
    public GameTheoryGUI() {
        initComponents();
        setLocation(20, 50);
        equilibriaTable.setEnabled(false);
        buttonGroupTableView.add(radioButtonMonteCarloData);
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    // <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:initComponents
    private void initComponents() {
        buttonGroupTableView = new javax.swing.ButtonGroup();
        buttonGroupMinimizerType = new javax.swing.ButtonGroup();
        TextPanel = new javax.swing.JPanel();
        jScrollPane4 = new javax.swing.JScrollPane();
        jt = new javax.swing.JTextArea();
        panelControl = new javax.swing.JPanel();
        buttonExecuteMinimization = new javax.swing.JButton();
        jLabel4 = new javax.swing.JLabel();
        importanceBidCostField = new javax.swing.JTextField();
        jLabel5 = new javax.swing.JLabel();
        monteCarloButton = new javax.swing.JButton();
        jLabel7 = new javax.swing.JLabel();
        jLabel8 = new javax.swing.JLabel();
        spinnerNumPlayers = new javax.swing.JSpinner();
        graphPanel = new javax.swing.JPanel();
        spinnerMonteCarloNumObs = new javax.swing.JSpinner();
        jLabel1 = new javax.swing.JLabel();
        spinnerMonteCarloNumberIterations = new javax.swing.JSpinner();
        jLabel6 = new javax.swing.JLabel();
        textFieldPureShifter = new javax.swing.JTextField();
        jLabel9 = new javax.swing.JLabel();
        fieldEfficientShifter = new javax.swing.JTextField();
        jLabel10 = new javax.swing.JLabel();
        spinnerParameterPlot = new javax.swing.JSpinner();
        jLabel11 = new javax.swing.JLabel();
        spinnerOutcomesPerGame = new javax.swing.JSpinner();
        checkBoxPlot = new javax.swing.JCheckBox();
        buttonGenerateImportanceGames = new javax.swing.JButton();
        buttonLoadData = new javax.swing.JButton();
        buttonPlotFunction = new javax.swing.JButton();
        buttonStartingValues = new javax.swing.JButton();
        jLabel12 = new javax.swing.JLabel();
        fieldProfitScaleFactor = new javax.swing.JTextField();
        spinnerNumberImportanceGames = new javax.swing.JSpinner();
        jLabel13 = new javax.swing.JLabel();
        spinnerNumberThreads = new javax.swing.JSpinner();
        BarPanel = new javax.swing.JPanel();
        bar = new javax.swing.JProgressBar();
        labelStatus = new javax.swing.JLabel();
        jPanel1 = new javax.swing.JPanel();
        jPanel2 = new javax.swing.JPanel();
        jLabel2 = new javax.swing.JLabel();
        jSpinner1 = new javax.swing.JSpinner();
        jRadioButton1 = new javax.swing.JRadioButton();
        jRadioButton2 = new javax.swing.JRadioButton();
        radioButtonMonteCarloData = new javax.swing.JRadioButton();
        jLabel3 = new javax.swing.JLabel();
        jLabel14 = new javax.swing.JLabel();
        jLabel15 = new javax.swing.JLabel();
        jLabel17 = new javax.swing.JLabel();
        tabbedPaneTable = new javax.swing.JTabbedPane();
        panelPayoffs = new javax.swing.JPanel();
        jScrollPane1 = new javax.swing.JScrollPane();
        jTable1 = new javax.swing.JTable();
        panelEquilibria = new javax.swing.JPanel();
        jScrollPane2 = new javax.swing.JScrollPane();
        equilibriaTable = new javax.swing.JTable();
        panelEstimates = new javax.swing.JPanel();
        jScrollPane3 = new javax.swing.JScrollPane();
        tableResults = new javax.swing.JTable();

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setTitle("Game Theory Estimation Control Panel");
        TextPanel.setBorder(javax.swing.BorderFactory.createTitledBorder("Text Output"));
        jt.setColumns(20);
        jt.setEditable(false);
        jt.setRows(5);
        jScrollPane4.setViewportView(jt);

        org.jdesktop.layout.GroupLayout TextPanelLayout = new org.jdesktop.layout.GroupLayout(TextPanel);
        TextPanel.setLayout(TextPanelLayout);
        TextPanelLayout.setHorizontalGroup(
            TextPanelLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(TextPanelLayout.createSequentialGroup()
                .addContainerGap()
                .add(jScrollPane4, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 1379, Short.MAX_VALUE)
                .addContainerGap())
        );
        TextPanelLayout.setVerticalGroup(
            TextPanelLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(TextPanelLayout.createSequentialGroup()
                .add(jScrollPane4, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 146, Short.MAX_VALUE)
                .addContainerGap())
        );

        panelControl.setBorder(javax.swing.BorderFactory.createTitledBorder("Control Panel"));
        buttonExecuteMinimization.setText("Execute Minimization");
        buttonExecuteMinimization.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                buttonExecuteMinimizationActionPerformed(evt);
            }
        });

        jLabel4.setText("Bid Cost:");

        importanceBidCostField.setHorizontalAlignment(javax.swing.JTextField.RIGHT);
        importanceBidCostField.setText(GameTheoryConstants.BID_COST1+"");
        importanceBidCostField.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                importanceBidCostFieldActionPerformed(evt);
            }
        });

        jLabel5.setText("Number of Importance Games:");

        monteCarloButton.setText("Execute Monte Carlo");
        monteCarloButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                monteCarloButtonActionPerformed(evt);
            }
        });

        jLabel7.setText("Monte Carlo NumObs:");

        jLabel8.setText("Number of Players:");

        spinnerNumPlayers.setModel(new SpinnerNumberModel(GameTheoryConstants.NUM_PLAYERS, 2, 5, 1));
        spinnerNumPlayers.setValue(GameTheoryConstants.NUM_PLAYERS);
        spinnerNumPlayers.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                spinnerNumPlayersStateChanged(evt);
            }
        });

        graphPanel.setLayout(new java.awt.BorderLayout());

        graphPanel.setBorder(javax.swing.BorderFactory.createTitledBorder("Plot: Objective Function"));

        spinnerMonteCarloNumObs.setModel(new SpinnerNumberModel(GameTheoryConstants.NUMBER_MONTE_CARLO_GAMES, GameTheoryConstants.NUM_PARAMS, 500, 1));
        spinnerMonteCarloNumObs.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                spinnerMonteCarloNumObsStateChanged(evt);
            }
        });

        jLabel1.setText("Number of Monte Carlos:");

        spinnerMonteCarloNumberIterations.setModel(new SpinnerNumberModel(GameTheoryConstants.NUMBER_MONTE_CARLO_ITERATIONS, 1, 500, 1));
        spinnerMonteCarloNumberIterations.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                spinnerMonteCarloNumberIterationsStateChanged(evt);
            }
        });

        jLabel6.setText("Pure Shifter:");

        textFieldPureShifter.setHorizontalAlignment(javax.swing.JTextField.RIGHT);
        textFieldPureShifter.setText(GameTheoryConstants.PURE_STRATEGY_EQUILIBRIUM_SHIFTER+"");
        textFieldPureShifter.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                textFieldPureShifterActionPerformed(evt);
            }
        });

        jLabel9.setText("Efficient Shifter:");

        fieldEfficientShifter.setHorizontalAlignment(javax.swing.JTextField.RIGHT);
        fieldEfficientShifter.setText(GameTheoryConstants.EFFICIENT_EQUILIBRIUM_SHIFTER+"");
        fieldEfficientShifter.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                fieldEfficientShifterActionPerformed(evt);
            }
        });

        jLabel10.setText("Parameter to Plot:");

        spinnerParameterPlot.setModel(new SpinnerNumberModel(GameTheoryConstants.PLOT_PARAMETER, 1, GameTheoryConstants.NUM_PARAMS, 1));
        spinnerParameterPlot.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                spinnerParameterPlotStateChanged(evt);
            }
        });

        jLabel11.setText("Outcomes per Game:");

        spinnerOutcomesPerGame.setModel(new SpinnerNumberModel(GameTheoryConstants.NUMBER_OUTCOMES_PER_GAME, 1, 50, 1));
        spinnerOutcomesPerGame.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                spinnerOutcomesPerGameStateChanged(evt);
            }
        });

        checkBoxPlot.setText("Plot");
        checkBoxPlot.setBorder(javax.swing.BorderFactory.createEmptyBorder(0, 0, 0, 0));
        checkBoxPlot.setHorizontalAlignment(javax.swing.SwingConstants.LEFT);
        checkBoxPlot.setMargin(new java.awt.Insets(0, 0, 0, 0));
        checkBoxPlot.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                checkBoxPlotActionPerformed(evt);
            }
        });

        buttonGenerateImportanceGames.setText("Generate Importance Games");
        buttonGenerateImportanceGames.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                buttonGenerateImportanceGamesActionPerformed(evt);
            }
        });

        buttonLoadData.setText("Load Data");
        buttonLoadData.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                buttonLoadDataActionPerformed(evt);
            }
        });

        buttonPlotFunction.setText("Plot Function");
        buttonPlotFunction.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                buttonPlotFunctionActionPerformed(evt);
            }
        });

        buttonStartingValues.setText("Get Starting Values");
        buttonStartingValues.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                buttonStartingValuesActionPerformed(evt);
            }
        });

        jLabel12.setText("Profit Scale Factor:");

        fieldProfitScaleFactor.setHorizontalAlignment(javax.swing.JTextField.RIGHT);
        fieldProfitScaleFactor.setText(GameTheoryConstants.PROFIT_MULTIPLIER+"");

        spinnerNumberImportanceGames.setModel(new SpinnerNumberModel(GameTheoryConstants.NUMBER_IMPORTANCE_GAMES, 1, 500, 1));
        spinnerNumberImportanceGames.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                spinnerNumberImportanceGamesStateChanged(evt);
            }
        });

        jLabel13.setText("Number of Processors:");

        spinnerNumberThreads.setModel(new SpinnerNumberModel(GameTheoryConstants.NUMBER_THREADS, 1, 2, 1));
        spinnerNumberThreads.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                spinnerNumberThreadsStateChanged(evt);
            }
        });

        org.jdesktop.layout.GroupLayout panelControlLayout = new org.jdesktop.layout.GroupLayout(panelControl);
        panelControl.setLayout(panelControlLayout);
        panelControlLayout.setHorizontalGroup(
            panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(org.jdesktop.layout.GroupLayout.TRAILING, panelControlLayout.createSequentialGroup()
                .addContainerGap()
                .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.TRAILING)
                    .add(org.jdesktop.layout.GroupLayout.LEADING, graphPanel, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 427, Short.MAX_VALUE)
                    .add(org.jdesktop.layout.GroupLayout.LEADING, panelControlLayout.createSequentialGroup()
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                            .add(jLabel4)
                            .add(jLabel5)
                            .add(jLabel8)
                            .add(jLabel7)
                            .add(monteCarloButton))
                        .add(10, 10, 10)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                            .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING, false)
                                .add(spinnerNumPlayers)
                                .add(importanceBidCostField, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 60, Short.MAX_VALUE)
                                .add(spinnerMonteCarloNumObs)
                                .add(spinnerNumberImportanceGames))
                            .add(checkBoxPlot, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 70, Short.MAX_VALUE))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                            .add(jLabel10)
                            .add(jLabel9)
                            .add(jLabel6)
                            .add(jLabel1)
                            .add(jLabel11)
                            .add(jLabel12))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING, false)
                            .add(fieldProfitScaleFactor)
                            .add(spinnerOutcomesPerGame)
                            .add(spinnerParameterPlot)
                            .add(fieldEfficientShifter)
                            .add(org.jdesktop.layout.GroupLayout.TRAILING, textFieldPureShifter)
                            .add(org.jdesktop.layout.GroupLayout.TRAILING, spinnerMonteCarloNumberIterations, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, 56, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)))
                    .add(org.jdesktop.layout.GroupLayout.LEADING, panelControlLayout.createSequentialGroup()
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.TRAILING, false)
                            .add(buttonLoadData, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                            .add(buttonPlotFunction, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING, false)
                            .add(buttonStartingValues, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                            .add(buttonGenerateImportanceGames, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                            .add(buttonExecuteMinimization, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 149, Short.MAX_VALUE)
                            .add(panelControlLayout.createSequentialGroup()
                                .add(jLabel13)
                                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                                .add(spinnerNumberThreads, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 36, Short.MAX_VALUE)))))
                .addContainerGap())
        );
        panelControlLayout.setVerticalGroup(
            panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(panelControlLayout.createSequentialGroup()
                .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                    .add(buttonGenerateImportanceGames)
                    .add(buttonLoadData)
                    .add(buttonExecuteMinimization))
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                    .add(buttonPlotFunction)
                    .add(buttonStartingValues)
                    .add(jLabel13)
                    .add(spinnerNumberThreads, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                    .add(panelControlLayout.createSequentialGroup()
                        .add(6, 6, 6)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                            .add(jLabel4)
                            .add(importanceBidCostField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                            .add(jLabel5)
                            .add(spinnerNumberImportanceGames, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                        .add(27, 27, 27)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                            .add(jLabel8)
                            .add(spinnerNumPlayers, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                            .add(jLabel7)
                            .add(spinnerMonteCarloNumObs, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                            .add(monteCarloButton)
                            .add(checkBoxPlot)))
                    .add(panelControlLayout.createSequentialGroup()
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                            .add(jLabel6)
                            .add(textFieldPureShifter, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                            .add(jLabel9)
                            .add(fieldEfficientShifter, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                            .add(jLabel12)
                            .add(fieldProfitScaleFactor, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                            .add(jLabel1)
                            .add(spinnerMonteCarloNumberIterations, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                            .add(jLabel10)
                            .add(spinnerParameterPlot, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(panelControlLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                            .add(jLabel11)
                            .add(spinnerOutcomesPerGame, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))))
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(graphPanel, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 427, Short.MAX_VALUE)
                .addContainerGap())
        );

        BarPanel.setBorder(javax.swing.BorderFactory.createTitledBorder("Monitor"));

        labelStatus.setText("Status: Idle");

        org.jdesktop.layout.GroupLayout BarPanelLayout = new org.jdesktop.layout.GroupLayout(BarPanel);
        BarPanel.setLayout(BarPanelLayout);
        BarPanelLayout.setHorizontalGroup(
            BarPanelLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(org.jdesktop.layout.GroupLayout.TRAILING, BarPanelLayout.createSequentialGroup()
                .addContainerGap()
                .add(labelStatus)
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED, 1223, Short.MAX_VALUE)
                .add(bar, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                .addContainerGap())
        );
        BarPanelLayout.setVerticalGroup(
            BarPanelLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(BarPanelLayout.createSequentialGroup()
                .add(BarPanelLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                    .add(bar, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                    .add(labelStatus))
                .addContainerGap(org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
        );

        jPanel1.setBorder(javax.swing.BorderFactory.createTitledBorder("Results"));
        jPanel2.setBorder(javax.swing.BorderFactory.createTitledBorder("Information"));
        jLabel2.setText("Game:");

        jRadioButton1.setText("Data");
        jRadioButton1.setBorder(javax.swing.BorderFactory.createEmptyBorder(0, 0, 0, 0));
        jRadioButton1.setMargin(new java.awt.Insets(0, 0, 0, 0));

        jRadioButton2.setText("Importance");
        jRadioButton2.setBorder(javax.swing.BorderFactory.createEmptyBorder(0, 0, 0, 0));
        jRadioButton2.setMargin(new java.awt.Insets(0, 0, 0, 0));

        radioButtonMonteCarloData.setText("Monte Carlo");
        radioButtonMonteCarloData.setBorder(javax.swing.BorderFactory.createEmptyBorder(0, 0, 0, 0));
        radioButtonMonteCarloData.setMargin(new java.awt.Insets(0, 0, 0, 0));

        jLabel3.setText("# Equilibria:");

        jLabel14.setText("Efficient Index:");

        jLabel15.setText("Outcome:");

        jLabel17.setText("# Mixed:");

        org.jdesktop.layout.GroupLayout jPanel2Layout = new org.jdesktop.layout.GroupLayout(jPanel2);
        jPanel2.setLayout(jPanel2Layout);
        jPanel2Layout.setHorizontalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(jPanel2Layout.createSequentialGroup()
                .addContainerGap()
                .add(jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                    .add(jPanel2Layout.createSequentialGroup()
                        .add(jLabel2)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(jSpinner1, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, 39, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(jRadioButton1)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(jRadioButton2)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(radioButtonMonteCarloData))
                    .add(jLabel17)
                    .add(jLabel14)
                    .add(jLabel15)
                    .add(jLabel3))
                .addContainerGap(603, Short.MAX_VALUE))
        );
        jPanel2Layout.setVerticalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(jPanel2Layout.createSequentialGroup()
                .add(jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                    .add(jLabel2)
                    .add(jSpinner1, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                    .add(jRadioButton1)
                    .add(jRadioButton2)
                    .add(radioButtonMonteCarloData))
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(jLabel3)
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(jLabel17)
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(jLabel14)
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(jLabel15)
                .addContainerGap(43, Short.MAX_VALUE))
        );

        jTable1.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {
                {null, null, null, null},
                {null, null, null, null},
                {null, null, null, null},
                {null, null, null, null}
            },
            new String [] {
                "Title 1", "Title 2", "Title 3", "Title 4"
            }
        ));
        jScrollPane1.setViewportView(jTable1);

        org.jdesktop.layout.GroupLayout panelPayoffsLayout = new org.jdesktop.layout.GroupLayout(panelPayoffs);
        panelPayoffs.setLayout(panelPayoffsLayout);
        panelPayoffsLayout.setHorizontalGroup(
            panelPayoffsLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(org.jdesktop.layout.GroupLayout.TRAILING, panelPayoffsLayout.createSequentialGroup()
                .addContainerGap()
                .add(jScrollPane1, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 885, Short.MAX_VALUE)
                .addContainerGap())
        );
        panelPayoffsLayout.setVerticalGroup(
            panelPayoffsLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(org.jdesktop.layout.GroupLayout.TRAILING, panelPayoffsLayout.createSequentialGroup()
                .addContainerGap()
                .add(jScrollPane1, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 408, Short.MAX_VALUE)
                .addContainerGap())
        );
        tabbedPaneTable.addTab("Payoffs", panelPayoffs);

        equilibriaTable.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {
                {null, null, null, null},
                {null, null, null, null},
                {null, null, null, null},
                {null, null, null, null}
            },
            new String [] {
                "Title 1", "Title 2", "Title 3", "Title 4"
            }
        ));
        jScrollPane2.setViewportView(equilibriaTable);

        org.jdesktop.layout.GroupLayout panelEquilibriaLayout = new org.jdesktop.layout.GroupLayout(panelEquilibria);
        panelEquilibria.setLayout(panelEquilibriaLayout);
        panelEquilibriaLayout.setHorizontalGroup(
            panelEquilibriaLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(org.jdesktop.layout.GroupLayout.TRAILING, panelEquilibriaLayout.createSequentialGroup()
                .addContainerGap()
                .add(jScrollPane2, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 885, Short.MAX_VALUE)
                .addContainerGap())
        );
        panelEquilibriaLayout.setVerticalGroup(
            panelEquilibriaLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(org.jdesktop.layout.GroupLayout.TRAILING, panelEquilibriaLayout.createSequentialGroup()
                .addContainerGap()
                .add(jScrollPane2, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 408, Short.MAX_VALUE)
                .addContainerGap())
        );
        tabbedPaneTable.addTab("Equilibria", panelEquilibria);

        tableResults.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {
                {null, null, null, null},
                {null, null, null, null},
                {null, null, null, null},
                {null, null, null, null}
            },
            new String [] {
                "Title 1", "Title 2", "Title 3", "Title 4"
            }
        ));
        jScrollPane3.setViewportView(tableResults);

        org.jdesktop.layout.GroupLayout panelEstimatesLayout = new org.jdesktop.layout.GroupLayout(panelEstimates);
        panelEstimates.setLayout(panelEstimatesLayout);
        panelEstimatesLayout.setHorizontalGroup(
            panelEstimatesLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(org.jdesktop.layout.GroupLayout.TRAILING, panelEstimatesLayout.createSequentialGroup()
                .addContainerGap()
                .add(jScrollPane3, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 885, Short.MAX_VALUE)
                .addContainerGap())
        );
        panelEstimatesLayout.setVerticalGroup(
            panelEstimatesLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(org.jdesktop.layout.GroupLayout.TRAILING, panelEstimatesLayout.createSequentialGroup()
                .addContainerGap()
                .add(jScrollPane3, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 408, Short.MAX_VALUE)
                .addContainerGap())
        );
        tabbedPaneTable.addTab("Estimates", panelEstimates);

        org.jdesktop.layout.GroupLayout jPanel1Layout = new org.jdesktop.layout.GroupLayout(jPanel1);
        jPanel1.setLayout(jPanel1Layout);
        jPanel1Layout.setHorizontalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(org.jdesktop.layout.GroupLayout.TRAILING, jPanel1Layout.createSequentialGroup()
                .addContainerGap()
                .add(jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.TRAILING)
                    .add(org.jdesktop.layout.GroupLayout.LEADING, tabbedPaneTable, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 910, Short.MAX_VALUE)
                    .add(org.jdesktop.layout.GroupLayout.LEADING, jPanel2, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
                .addContainerGap())
        );
        jPanel1Layout.setVerticalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(jPanel1Layout.createSequentialGroup()
                .add(jPanel2, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(tabbedPaneTable, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 458, Short.MAX_VALUE)
                .addContainerGap())
        );

        org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(layout.createSequentialGroup()
                .addContainerGap()
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                    .add(TextPanel, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                    .add(layout.createSequentialGroup()
                        .add(panelControl, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(jPanel1, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
                    .add(BarPanel, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
                .addContainerGap())
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(layout.createSequentialGroup()
                .addContainerGap()
                .add(TextPanel, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING, false)
                    .add(panelControl, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                    .add(jPanel1, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(BarPanel, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, 48, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                .addContainerGap(org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
        );
        pack();
    }// </editor-fold>//GEN-END:initComponents

    /**
     * Updates the number of threads based on a change in the associated JSpinner state.
     * @param evt GUI event.
     */
    private void spinnerNumberThreadsStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_spinnerNumberThreadsStateChanged
        GameTheoryConstants.NUMBER_THREADS = (Integer) spinnerNumberThreads.getValue();
    }//GEN-LAST:event_spinnerNumberThreadsStateChanged

    /**
     * Updates the number of importance games based on a change in the associated JSpinner state.
     * @param evt GUI event.
     */
    private void spinnerNumberImportanceGamesStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_spinnerNumberImportanceGamesStateChanged
        GameTheoryConstants.NUMBER_IMPORTANCE_GAMES = (Integer) spinnerNumberImportanceGames.getValue();
    }//GEN-LAST:event_spinnerNumberImportanceGamesStateChanged

    /**
     * Method to obtain starting values by estimating a private information analogue of the auction entry game.
     *
     * @param evt Mouse button event.
     */
    private void buttonStartingValuesActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_buttonStartingValuesActionPerformed
        append("Obtaining starting values...");
        setButtonsBusy();

        SwingWorker worker = new SwingWorker() {

            NumberFormat nf = NumberFormat.getInstance();
            Jama.Matrix results;

            public void finished() {
                append("OK\n" + pmUtility.stringPrettyPrint(results) + "\n");
                GameTheoryConstants.BID_COST1 = results.get(0, 1);
                GameTheoryConstants.BID_COST2 = results.get(0, 1);
                GameTheoryConstants.BID_COST3 = results.get(0, 1);
                GameTheoryConstants.BID_COST4 = results.get(0, 1);
                GameTheoryConstants.PROFIT_MULTIPLIER = results.get(0, 2);
                fieldProfitScaleFactor.setText(nf.format(GameTheoryConstants.PROFIT_MULTIPLIER));
                importanceBidCostField.setText(nf.format(GameTheoryConstants.BID_COST1));
                setButtonsReady();
            }

            @Override
            protected Object doInBackground() {
                GameMinimizationPrivateInformation go = new GameMinimizationPrivateInformation(gameList);
                results = new Jama.Matrix(go.execute(), 1);
                return null;
            }
        };
        worker.execute();

    }//GEN-LAST:event_buttonStartingValuesActionPerformed
    /**
     * Generate a graph.
     *
     * @param evt Mouse click event.
     */
    private void buttonPlotFunctionActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_buttonPlotFunctionActionPerformed
        graph();
    }//GEN-LAST:event_buttonPlotFunctionActionPerformed

    /**
     * Generate importance games.
     *
     * @param evt Mouse click event.
     */
    private void buttonGenerateImportanceGamesActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_buttonGenerateImportanceGamesActionPerformed
        generateImportanceGames();
    }//GEN-LAST:event_buttonGenerateImportanceGamesActionPerformed

    /**
     * Load in the data.
     *
     * @param evt Mouse click event.
     */
    private void buttonLoadDataActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_buttonLoadDataActionPerformed
        loadData();
    }//GEN-LAST:event_buttonLoadDataActionPerformed

    /**
     * Update a constant based on a change in the underlying GUI component.
     * @param evt A change of state event.
     */
    private void checkBoxPlotActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_checkBoxPlotActionPerformed
        GameTheoryConstants.SHOW_PLOT = checkBoxPlot.isSelected();
    }//GEN-LAST:event_checkBoxPlotActionPerformed
    /**
     * Update a constant based on a change in the underlying GUI component.
     * @param evt A change of state event.
     */
    private void spinnerOutcomesPerGameStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_spinnerOutcomesPerGameStateChanged
        GameTheoryConstants.NUMBER_OUTCOMES_PER_GAME = (Integer) spinnerOutcomesPerGame.getValue();
    }//GEN-LAST:event_spinnerOutcomesPerGameStateChanged
    /**
     * Update a constant based on a change in the underlying GUI component.
     * @param evt A change of state event.
     */
    private void spinnerParameterPlotStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_spinnerParameterPlotStateChanged
        GameTheoryConstants.PLOT_PARAMETER = (Integer) spinnerParameterPlot.getValue();
    }//GEN-LAST:event_spinnerParameterPlotStateChanged
    /**
     * Update a constant based on a change in the underlying GUI component.
     * @param evt A change of state event.
     */
    private void fieldEfficientShifterActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_fieldEfficientShifterActionPerformed
        GameTheoryConstants.EFFICIENT_EQUILIBRIUM_SHIFTER = Double.parseDouble(fieldEfficientShifter.getText());
    }//GEN-LAST:event_fieldEfficientShifterActionPerformed
    /**
     * Update a constant based on a change in the underlying GUI component.
     * @param evt A change of state event.
     */
    private void textFieldPureShifterActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_textFieldPureShifterActionPerformed
        GameTheoryConstants.PURE_STRATEGY_EQUILIBRIUM_SHIFTER = Double.parseDouble(textFieldPureShifter.getText());
    }//GEN-LAST:event_textFieldPureShifterActionPerformed
    /**
     * Update a constant based on a change in the underlying GUI component.
     * @param evt A change of state event.
     */
    private void spinnerMonteCarloNumberIterationsStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_spinnerMonteCarloNumberIterationsStateChanged
        GameTheoryConstants.NUMBER_MONTE_CARLO_ITERATIONS = (Integer) spinnerMonteCarloNumberIterations.getValue();
    }//GEN-LAST:event_spinnerMonteCarloNumberIterationsStateChanged
    /**
     * Update a constant based on a change in the underlying GUI component.
     * @param evt A change of state event.
     */
    private void spinnerMonteCarloNumObsStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_spinnerMonteCarloNumObsStateChanged
        GameTheoryConstants.NUMBER_MONTE_CARLO_GAMES = (Integer) spinnerMonteCarloNumObs.getValue();
    }//GEN-LAST:event_spinnerMonteCarloNumObsStateChanged
    /**
     * Update several constants based on a change in the underlying GUI component.
     * @param evt A change of state event.
     */
    private void spinnerNumPlayersStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_spinnerNumPlayersStateChanged
        GameTheoryConstants.NUM_PLAYERS = (Integer) spinnerNumPlayers.getValue();
        GameTheoryConstants.NUM_PARAMS = (Integer) spinnerNumPlayers.getValue() + 4;
        GenerateRecursiveList lister = new GenerateRecursiveList(GameTheoryConstants.NUM_PLAYERS, 2);
        GameTheoryConstants.actionList = lister.getCombinations();
    }//GEN-LAST:event_spinnerNumPlayersStateChanged

    /**
     * Run a Monte Carlo.  Note that this was a residual of a debugging process, to ensure that our estimator
     * was programmed correctly.  The Monte Carlo results in the paper come from a dedicated Monte Carlo
     * program, which is also included with this source code.  This code fragment is left here for completeness.
     *
     * @param evt Mouse click event.
     */
    private void monteCarloButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_monteCarloButtonActionPerformed
        // this is to execute the Monte Carlo
        // what do we need to do to run a monte carlo?
        // need to generate the data (payoffs, outcomes)
        // need to generate importance games and solve their equilibria
        // need to load both of those into a coherent data object and then run the minimizer
        // then report the results
        // the basic structure of this program is that gameList holds the list of all games
        // and importanceGamesList holds all the importance games
        // true parameters held in GTConstants

        setButtonsBusy();

        SwingWorker worker = new SwingWorker() {

            Jama.Matrix results = new Jama.Matrix(GameTheoryConstants.NUMBER_MONTE_CARLO_ITERATIONS, GameTheoryConstants.NUM_PARAMS + 1);

            @Override
            protected Object doInBackground() {
                try {
                    BufferedWriter out = new BufferedWriter(new FileWriter("montecarlo.txt", false));
                    out.close();
                } catch (Exception e) {
                    e.printStackTrace();
                }

                for (GameTheoryConstants.NUMBER_MONTE_CARLO_GAMES = 100; GameTheoryConstants.NUMBER_MONTE_CARLO_GAMES <= 400; GameTheoryConstants.NUMBER_MONTE_CARLO_GAMES *= 2) {
                    // for(int OLP=0;OLP<1;OLP++) {
                    labelStatus.setText("Status: Running Monte Carlo");
                    labelStatus.setText("Status: Generating games and solving equilibria...");
                    // use the importance games to generate outcomes, then see if we can turn around and match the unknown parameters given the X's and outcomes
                    // so what do i have to do here, exactly?
                    // take each of the data points, and

                    int lineNumber = 0;

                    try {
                        BufferedReader in = new BufferedReader(new FileReader("results.csv"));
                        String line = in.readLine();
                        while (line != null) {
                            if (line != null) {
                                int a = 0;
                                int b = line.indexOf(",", a);
                                for (int i = 0; i < 7; i++) {
                                    a = b + 1;
                                    b = line.indexOf(",", a);
                                    results.set(lineNumber, i + 1, Double.parseDouble(line.substring(a, b)));
                                }
                                // a = b+1;
                                // results.set(lineNumber, 7, Double.parseDouble(line.substring(a)));
                                lineNumber++;
                                line = in.readLine();
                            }
                        }
                        in.close();
                    } catch (Exception e) {
                        e.printStackTrace();
                    }

                    // pmUtility.prettyPrint(results);
                    // System.exit(0);

                    for (int monte = lineNumber; monte < GameTheoryConstants.NUMBER_MONTE_CARLO_ITERATIONS; monte++) {
                        setBarMaximum(GameTheoryConstants.NUMBER_MONTE_CARLO_GAMES);
                        monteCarloData = new ArrayList<Game>();
                        // add games to monteCarloData using threaded list
                        labelStatus.setText("Status: Generating data for Monte Carlo...(" + (monte + 1) + "/" + GameTheoryConstants.NUMBER_MONTE_CARLO_ITERATIONS + ")");

                        try {
                            ArrayList<Future<Game>> workerList = new ArrayList<Future<Game>>();
                            for (int i = 0; i < GameTheoryConstants.NUMBER_MONTE_CARLO_GAMES; i++) {
                                workerList.add(tpes.submit(new GenerateMonteCarloTask(i)));
                            }
                            for (int i = 0; i < GameTheoryConstants.NUMBER_MONTE_CARLO_GAMES; i++) {
                                Game g = workerList.get(i).get();
                                setBarValue(i + 1);
                                g.getSimulatedOutcome();
                                monteCarloData.add(g);
                            }
                        } catch (Exception e) {
                            e.printStackTrace();
                        }

                        labelStatus.setText("Status: Generating importance games for Monte Carlo...(" + (monte + 1) + "/" + GameTheoryConstants.NUMBER_MONTE_CARLO_ITERATIONS + ")");
                        importanceGamesList = new ArrayList<Game>();

                        try {
                            ArrayList<Future<Game>> workerList = new ArrayList<Future<Game>>();
                            for (int i = 0; i < GameTheoryConstants.NUMBER_MONTE_CARLO_GAMES; i++) {
                                workerList.add(tpes.submit(new GenerateMonteCarloTask(i)));
                            }

                            for (int i = 0; i < GameTheoryConstants.NUMBER_MONTE_CARLO_GAMES; i++) {
                                importanceGamesList.add(workerList.get(i).get());
                                setBarValue(i + 1);
                            }
                        } catch (Exception e) {
                            e.printStackTrace();
                        }

                        for (int i = 0; i < monteCarloData.size(); i++) {
                            (monteCarloData.get(i)).setImportanceGamesList(importanceGamesList);
                        }

                        GameMinimization go = new GameMinimization(monteCarloData);

                        double[] guess = setGuess();

                        double leftEndPoint = guess[GameTheoryConstants.PLOT_PARAMETER] - Math.max(1, Math.abs(guess[GameTheoryConstants.PLOT_PARAMETER]));
                        double rightEndPoint = guess[GameTheoryConstants.PLOT_PARAMETER] + Math.max(1, Math.abs(guess[GameTheoryConstants.PLOT_PARAMETER]));
                        double increment = (rightEndPoint - leftEndPoint) / 25.0;

                        if (GameTheoryConstants.SHOW_PLOT) {
                            labelStatus.setText("Status: Plotting graph");
                            XYSeries xy = new XYSeries("x[" + GameTheoryConstants.PLOT_PARAMETER + "]");
                            for (guess[GameTheoryConstants.PLOT_PARAMETER] = leftEndPoint; guess[GameTheoryConstants.PLOT_PARAMETER] <= rightEndPoint; guess[GameTheoryConstants.PLOT_PARAMETER] += increment) {
                                xy.add(guess[GameTheoryConstants.PLOT_PARAMETER], go.f_to_minimize(guess));
                            }
                            XYSeriesCollection collection = new XYSeriesCollection(xy);
                            ChartPanel panel = new ChartPanel(ChartFactory.createXYLineChart(null, "x[" + GameTheoryConstants.PLOT_PARAMETER + "]", "Objective Function", collection, PlotOrientation.VERTICAL, false, false, false));
                            graphPanel.removeAll();
                            graphPanel.add(panel);
                            graphPanel.validate();
                        }

                        guess = setGuess();

                        labelStatus.setText("Status: Calling minimizer...(" + (monte + 1) + "/" + GameTheoryConstants.NUMBER_MONTE_CARLO_ITERATIONS + ")");

                        setBarMaximum(GameTheoryConstants.NUMBER_MONTE_CARLO_ITERATIONS);
                        setBarValue(monte + 1);
                        Jama.Matrix iresults = new Jama.Matrix(go.execute(guess), 1);
                        // append("Results: "+pmUtility.stringPrettyPrint(iresults)+"\n");
                        // append("OK\n");
                        for (int p = 0; p < GameTheoryConstants.NUM_PARAMS; p++) {
                            results.set(monte, p + 1, iresults.get(0, p + 1));
                        }
                        try {
                            BufferedWriter out = new BufferedWriter(new FileWriter("results.csv", true));
                            for (int k = 0; k < results.getColumnDimension(); k++) {
                                out.write(results.get(monte, k) + ",");
                            }
                            out.write("\n");
                            out.close();
                        } catch (Exception e) {
                            e.printStackTrace();
                        }
                    }
                    outputMatrixResults(results);

                    // only when complete should i delete the file
                    try {
                        BufferedWriter out = new BufferedWriter(new FileWriter("results.csv", false));
                        out.close();
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                }
                return null;
            }

            @Override
            public void done() {
                setBarValue(0);
                labelStatus.setText("Status: Idle");
                setButtonsReady();
            }
        };
        worker.execute();
    }//GEN-LAST:event_monteCarloButtonActionPerformed

    /**
     * Set the initial guess at some parameters defined in GameTheoryConstants.
     *
     * @return An array containing the updated guess at the true parameters.
     */
    public double[] setGuess() {
        double[] guess = new double[GameTheoryConstants.NUM_PARAMS + 1];
        guess[1] = GameTheoryConstants.BID_COST1;
        guess[2] = GameTheoryConstants.PURE_STRATEGY_EQUILIBRIUM_SHIFTER;
        guess[3] = GameTheoryConstants.EFFICIENT_EQUILIBRIUM_SHIFTER;
        guess[4] = GameTheoryConstants.DOMINATED_EQUILIBRIUM_SHIFTER;
        guess[5] = GameTheoryConstants.PROFIT_MULTIPLIER;

        if (GameTheoryConstants.NUM_PLAYERS >= 2) {
            guess[6] = GameTheoryConstants.BID_COST2;
        }
        if (GameTheoryConstants.NUM_PLAYERS >= 3) {
            guess[7] = GameTheoryConstants.BID_COST3;
        }
        if (GameTheoryConstants.NUM_PLAYERS >= 4) {
            guess[8] = GameTheoryConstants.BID_COST4;
        }

        return guess;
    }

    /**
     * Method that graphs the GMM objective function with respect to one parameter.
     */
    public void graph() {
        SwingWorker worker = new SwingWorker() {

            Jama.Matrix results = new Jama.Matrix(GameTheoryConstants.NUMBER_MONTE_CARLO_ITERATIONS, GameTheoryConstants.NUM_PARAMS + 1);

            @Override
            protected Object doInBackground() {
                GameMinimization go = new GameMinimization(gameList);
                double[] guess = setGuess();

                double leftEndPoint = guess[GameTheoryConstants.PLOT_PARAMETER] - Math.max(1, Math.abs(guess[GameTheoryConstants.PLOT_PARAMETER]));
                double rightEndPoint = guess[GameTheoryConstants.PLOT_PARAMETER] + Math.max(1, Math.abs(guess[GameTheoryConstants.PLOT_PARAMETER]));
                double increment = (rightEndPoint - leftEndPoint) / 25.0;

                labelStatus.setText("Status: Plotting graph");
                XYSeries xy = new XYSeries("x[" + GameTheoryConstants.PLOT_PARAMETER + "]");
                setBarMaximum(25);
                int counter = 0;
                for (guess[GameTheoryConstants.PLOT_PARAMETER] = leftEndPoint; guess[GameTheoryConstants.PLOT_PARAMETER] <= rightEndPoint; guess[GameTheoryConstants.PLOT_PARAMETER] += increment) {
                    counter++;
                    setBarValue(counter);
                    // System.out.print("Sending: ");
                    // pmUtility.prettyPrint(new Jama.Matrix(guess, 1));
                    xy.add(guess[GameTheoryConstants.PLOT_PARAMETER], go.f_to_minimize(guess));
                }
                setBarValue(0);
                XYSeriesCollection collection = new XYSeriesCollection(xy);
                ChartPanel panel = new ChartPanel(ChartFactory.createXYLineChart(null, "x[" + GameTheoryConstants.PLOT_PARAMETER + "]", "Objective Function", collection, PlotOrientation.VERTICAL, false, false, false));
                graphPanel.removeAll();
                graphPanel.add(panel);
                graphPanel.revalidate();

                return null;
            }

            @Override
            public void done() {
                setBarValue(0);
                labelStatus.setText("Status: Idle");
                setButtonsReady();
            }
        };
        worker.execute();
    }

    /**
     * Update a constant based on a change in the underlying GUI component.
     * @param evt A change of state event.
     */
    private void importanceBidCostFieldActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_importanceBidCostFieldActionPerformed
        JTextField src = (JTextField) evt.getSource();
        GameTheoryConstants.BID_COST1 = Double.parseDouble(src.getText());
        append("Bid cost set to: " + GameTheoryConstants.BID_COST1 + "\n");
    }//GEN-LAST:event_importanceBidCostFieldActionPerformed

    /**
     * Optimizes the objective function.
     *
     * @param evt Mouse click event.
     */
    private void buttonExecuteMinimizationActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_buttonExecuteMinimizationActionPerformed
        GameTheoryConstants.NUM_PLAYERS = 4;
        setButtonsBusy();

        SwingWorker worker = new SwingWorker() {

            double[] guess;

            @Override
            protected Object doInBackground() {
                append("Running minimization...\n");
                labelStatus.setText("Status: Executing minimization...");
                guess = setGuess();

                GameMinimization go = new GameMinimization(gameList);
                if (GameTheoryConstants.USE_LTE) {
                    outputLTEResults(go.executeLTE(guess));
                } else {
                    guess = go.execute(guess);
                    append("Minimized at: " + pmUtility.stringPrettyPrint(new Jama.Matrix(guess, 1)) + "\n");
                    append("function value: " + go.f_to_minimize(guess) + "\n");
                }
                return null;
            }

            @Override
            public void done() {
                setButtonsReady();
                setBarValue(0);
                labelStatus.setText("Status: Idle");
                setConstants(guess);
            }
        };
        worker.execute();
    }//GEN-LAST:event_buttonExecuteMinimizationActionPerformed

    /**
     * Utility method to set parameters.
     *
     * @param guess Vector of parameters to set.
     */
    public void setConstants(double[] guess) {
        NumberFormat nf = NumberFormat.getInstance();

        GameTheoryConstants.BID_COST1 = guess[1];
        importanceBidCostField.setText(nf.format(guess[1]));

        if (GameTheoryConstants.NUM_PARAMS == 4) {
            GameTheoryConstants.PURE_STRATEGY_EQUILIBRIUM_SHIFTER = guess[2];
            GameTheoryConstants.EFFICIENT_EQUILIBRIUM_SHIFTER = guess[3];
            GameTheoryConstants.PROFIT_MULTIPLIER = guess[4];
            fieldEfficientShifter.setText(nf.format(guess[3]));
            fieldProfitScaleFactor.setText(nf.format(guess[4]));
            textFieldPureShifter.setText(nf.format(guess[2]));
        }
        if (GameTheoryConstants.NUM_PARAMS == 2) {
            GameTheoryConstants.PROFIT_MULTIPLIER = guess[2];
            fieldProfitScaleFactor.setText(nf.format(guess[2]));
        }
        if (GameTheoryConstants.NUM_PARAMS == 6) {
            GameTheoryConstants.PURE_STRATEGY_EQUILIBRIUM_SHIFTER = guess[2];
            GameTheoryConstants.EFFICIENT_EQUILIBRIUM_SHIFTER = guess[3];
            GameTheoryConstants.PROFIT_MULTIPLIER = guess[4];
            GameTheoryConstants.BID_COST2 = guess[5];
            GameTheoryConstants.BID_COST3 = guess[6];
            fieldEfficientShifter.setText(nf.format(guess[3]));
            fieldProfitScaleFactor.setText(nf.format(guess[4]));
            textFieldPureShifter.setText(nf.format(guess[2]));
        }
        if (GameTheoryConstants.NUM_PARAMS == 7) {
            GameTheoryConstants.PURE_STRATEGY_EQUILIBRIUM_SHIFTER = guess[2];
            GameTheoryConstants.EFFICIENT_EQUILIBRIUM_SHIFTER = guess[3];
            GameTheoryConstants.PROFIT_MULTIPLIER = guess[4];
            GameTheoryConstants.BID_COST2 = guess[5];
            GameTheoryConstants.BID_COST3 = guess[6];
            GameTheoryConstants.BID_COST4 = guess[7];
            fieldEfficientShifter.setText(nf.format(guess[3]));
            fieldProfitScaleFactor.setText(nf.format(guess[4]));
            textFieldPureShifter.setText(nf.format(guess[2]));
        }
    }

    /**
     * Outputs results from the Chernozhukov and Hong (2003) LTE procedure.
     *
     * @param results Matrix containing the MCMC chain generated by the LTE procedure.
     */
    public void outputLTEResults(Jama.Matrix results) {
        com.sharkysoft.printf.engine.RealFormatter nf = new com.sharkysoft.printf.engine.RealFormatter();
        nf.setMinRightDigits(4);
        nf.setMaxRightDigits(4);
        nf.setAlignment(com.sharkysoft.printf.engine.AlignmentMode.gpRight);
        nf.setFieldWidth(7);

        double[] beta = setGuess();

        // append("& & & Standard & Mean & Median & Pct Mean & Pct Median & & &  \\\\ \n");
        // append("Parameter & Mean & Median & Deviation & Bias & Bias & Bias & Bias & MSE & MAD & Median AD \\\\ \n");

        String[] label = {"Bid Cost1", "Pure Strategy", "Efficient", "Dominated", "Profit Scale", "Bid Cost2", "Bid Cost3", "Bid Cost4"};

        Object[][] data = new Object[GameTheoryConstants.NUM_PARAMS][6];
        String[] columnNames = {"Parameter", "Mean", "Median", "StdDev", "0.025", "0.975"};

        // pmUtility.prettyPrint(results);

        for (int j = 0; j < GameTheoryConstants.NUM_PARAMS; j++) {
            double mean = pmUtility.mean(results, j + 1);
            double median = pmUtility.median(results, j + 1);
            double stddev = pmUtility.standardDeviation(results, j + 1);
            double leftPt = pmUtility.percentile(results, j + 1, 0.025);
            double rightPt = pmUtility.percentile(results, j + 1, 0.975);
            double CILeft = leftPt; // mean-rightPt;
            double CIRight = rightPt; // mean-leftPt;

            data[j][0] = label[j];
            data[j][1] = nf.format(mean);
            data[j][2] = nf.format(median);
            data[j][3] = nf.format(stddev);
            data[j][4] = nf.format(CILeft);
            data[j][5] = nf.format(CIRight);

            append(label[j]);
            append("\t & " + nf.format(mean) + "\t & " + nf.format(median) + "\t & " + nf.format(stddev));
            append("\t & " + nf.format(CILeft) + "\t & " + nf.format(CIRight) + " \\\\ \n");
        }

        DefaultTableModel model = new DefaultTableModel(data, columnNames);
        tableResults.setModel(model);
    }

    /**
     * Outputs results from the Monte Carlo above.
     *
     * @param results Matrix of results from the internal Monte Carlo.
     */
    public void outputMatrixResults(Jama.Matrix results) {
        com.sharkysoft.printf.engine.RealFormatter nf = new com.sharkysoft.printf.engine.RealFormatter();
        nf.setMinRightDigits(4);
        nf.setMaxRightDigits(4);
        nf.setAlignment(com.sharkysoft.printf.engine.AlignmentMode.gpRight);
        nf.setFieldWidth(7);

        double[] beta = setGuess();

        String[] label = {"Bid Cost1", "Pure Strategy", "Efficient", "Dominated", "Profit Scale", "Bid Cost2", "Bid Cost3", "Bid Cost4"};

        Object[][] data = new Object[GameTheoryConstants.NUM_PARAMS][11];
        String[] columnNames = {"Parameter", "Mean", "Median", "StdDev", "MeanBias", "MedianBias", "PctMeanBias", "PctMedianBias", "MSE", "Mean AD", "MedianAD"};

        try {
            BufferedWriter out = new BufferedWriter(new FileWriter("montecarlo.txt", true));

            append("\\\\ \n n = " + GameTheoryConstants.NUMBER_MONTE_CARLO_GAMES + " \\\\ \n \\\\ \n");
            out.write("\\\\ \n n = " + GameTheoryConstants.NUMBER_MONTE_CARLO_GAMES + " \\\\ \n \\\\ \n");

            for (int j = 0; j < GameTheoryConstants.NUM_PARAMS; j++) {
                double mean = pmUtility.mean(results, j + 1);
                double median = pmUtility.median(results, j + 1);
                double stddev = pmUtility.standardDeviation(results, j + 1);
                double biasMean = mean - beta[j + 1];
                double biasMedian = median - beta[j + 1];
                double pctBiasMean = 100.0 * biasMean / beta[j + 1];
                double pctBiasMedian = 100.0 * biasMedian / beta[j + 1];
                double meanSqrError = 0;
                Jama.Matrix absoluteDeviations = new Jama.Matrix(results.getRowDimension(), 1);
                for (int k = 0; k < results.getRowDimension(); k++) {
                    absoluteDeviations.set(k, 0, Math.abs(results.get(k, j + 1) - beta[j + 1]));
                    meanSqrError += Math.pow(results.get(k, j + 1) - beta[j + 1], 2);
                }
                meanSqrError /= results.getRowDimension();
                double meanAD = pmUtility.mean(absoluteDeviations, 0);
                double medianAD = pmUtility.median(absoluteDeviations, 0);

                data[j][0] = label[j];
                data[j][1] = nf.format(mean);
                data[j][2] = nf.format(median);
                data[j][3] = nf.format(stddev);
                data[j][4] = nf.format(biasMean);
                data[j][5] = nf.format(biasMedian);
                data[j][6] = nf.format(pctBiasMean);
                data[j][7] = nf.format(pctBiasMedian);
                data[j][8] = nf.format(meanSqrError);
                data[j][9] = nf.format(meanAD);
                data[j][10] = nf.format(medianAD);

                append(label[j]);
                append("\t & " + nf.format(mean) + "\t & " + nf.format(median) + "\t & " + nf.format(stddev));
                append("\t & " + nf.format(biasMean) + "\t & " + nf.format(biasMedian));
                append("\t & " + nf.format(pctBiasMean) + "\t & " + nf.format(pctBiasMedian) + "\t & " + nf.format(meanSqrError) + "\t & " + nf.format(meanAD) + "\t & " + nf.format(medianAD) + " \\\\ \n");

                out.write(label[j]);
                out.write("\t & " + nf.format(mean) + "\t & " + nf.format(median) + "\t & " + nf.format(stddev));
                out.write("\t & " + nf.format(biasMean) + "\t & " + nf.format(biasMedian));
                out.write("\t & " + nf.format(pctBiasMean) + "\t & " + nf.format(pctBiasMedian) + "\t & " + nf.format(meanSqrError) + "\t & " + nf.format(meanAD) + "\t & " + nf.format(medianAD) + " \\\\ \n");
            }
            append("\n");
            out.write("\n");
            out.close();
        } catch (Exception e) {
            e.printStackTrace();
        }

        DefaultTableModel model = new DefaultTableModel(data, columnNames);
        tableResults.setModel(model);
        tabbedPaneTable.setSelectedIndex(2);
    }

    /**
     * Method to load in the auction entry data.
     */
    public void loadData() {
        setButtonsBusy();
        SwingWorker worker = new SwingWorker() {

            public Object doInBackground() {
                append("Loading data...");
                // so now i have a gameList, with the real games in it
                // next step is to attach outcomes to each of those games
                int gameIndex = 0;

                // read in data
                ReadInData dataProcessor = new ReadInData();
                gameList = dataProcessor.getGameList();

                try {
                    BufferedReader in = new BufferedReader(new FileReader("actualbids.csv"));
                    String line = in.readLine(); // headers
                    line = in.readLine();
                    ArrayList<Integer> bidderList = new ArrayList<Integer>();  // this is a local list to keep track of who is bidding on each job
                    String currentJobID = "";
                    while (line != null) {
                        if (line != null) {
                            int a = 0;
                            int b = line.indexOf(",", a);
                            String jobID = line.substring(a, b);
                            a = b + 1;
                            Integer bidderID = Integer.parseInt(line.substring(a));
                            if (!currentJobID.equals(jobID)) {
                                // new job id
                                // process the current bidderList, check to make sure that it is not entirely populated by major firms
                                // if so, drop it; if not, assign that outcome to the appropriate game
                                boolean fringe = true;
                                if (bidderList.size() == 2) {
                                    if (bidderList.contains(new Integer(75)) && bidderList.contains(new Integer(104))) {
                                        fringe = false;
                                    }
                                    if (bidderList.contains(new Integer(75)) && bidderList.contains(new Integer(135))) {
                                        fringe = false;
                                    }
                                    if (bidderList.contains(new Integer(75)) && bidderList.contains(new Integer(147))) {
                                        fringe = false;
                                    }
                                    if (bidderList.contains(new Integer(104)) && bidderList.contains(new Integer(135))) {
                                        fringe = false;
                                    }
                                    if (bidderList.contains(new Integer(104)) && bidderList.contains(new Integer(147))) {
                                        fringe = false;
                                    }
                                    if (bidderList.contains(new Integer(135)) && bidderList.contains(new Integer(147))) {
                                        fringe = false;
                                    }
                                }
                                if (bidderList.size() == 3) {
                                    if (bidderList.contains(new Integer(75)) && bidderList.contains(new Integer(104)) && bidderList.contains(new Integer(135))) {
                                        fringe = false;
                                    }
                                    if (bidderList.contains(new Integer(75)) && bidderList.contains(new Integer(104)) && bidderList.contains(new Integer(147))) {
                                        fringe = false;
                                    }
                                    if (bidderList.contains(new Integer(75)) && bidderList.contains(new Integer(135)) && bidderList.contains(new Integer(147))) {
                                        fringe = false;
                                    }
                                    if (bidderList.contains(new Integer(104)) && bidderList.contains(new Integer(135)) && bidderList.contains(new Integer(147))) {
                                        fringe = false;
                                    }
                                }
                                if (bidderList.size() == 4) {
                                    if (bidderList.contains(new Integer(75)) && bidderList.contains(new Integer(104)) && bidderList.contains(new Integer(135)) && bidderList.contains(new Integer(147))) {
                                        fringe = false;
                                    }
                                }
                                if (!fringe) {
                                    // System.out.println("Throwing out "+jobID);
                                    // this works
                                }
                                if (fringe) {
                                    // count as proper game and assign outcome to game in order
                                    Jama.Matrix outcome = new Jama.Matrix(4, 1);
                                    if (bidderList.contains(new Integer(75))) {
                                        outcome.set(0, 0, 1);
                                    }
                                    if (bidderList.contains(new Integer(104))) {
                                        outcome.set(1, 0, 1);
                                    }
                                    if (bidderList.contains(new Integer(135))) {
                                        outcome.set(2, 0, 1);
                                    }
                                    if (bidderList.contains(new Integer(147))) {
                                        outcome.set(3, 0, 1);
                                    }
                                    Game g = gameList.get(gameIndex);
                                    // pmUtility.prettyPrintVector(outcome);
                                    g.setOutcome(outcome);
                                    gameIndex++;
                                    // this seems to be working, assigning the outcomes to the appropriate games
                                }
                                bidderList = new ArrayList<Integer>();
                                currentJobID = jobID;
                                bidderList.add(bidderID);
                            } else {
                                bidderList.add(bidderID);
                            }
                            line = in.readLine();
                        }
                    }
                    in.close();
                } catch (Exception e) {
                    e.printStackTrace();
                }

                return null;
            }

            @Override
            public void done() {
                setButtonsReady();
                append("OK\n");
            }
        };
        worker.execute();
    }

    /**
     * Enables all the buttons in the GUI so that an action can be performed.
     */
    public void setButtonsReady() {
        buttonExecuteMinimization.setEnabled(true);
        buttonGenerateImportanceGames.setEnabled(true);
        buttonLoadData.setEnabled(true);
        monteCarloButton.setEnabled(true);
        buttonStartingValues.setEnabled(true);
        buttonPlotFunction.setEnabled(true);
    }

    /**
     * Disables the buttons while some task is processing in the background so that tasks are accomplished sequentially.
     */
    public void setButtonsBusy() {
        buttonExecuteMinimization.setEnabled(false);
        buttonGenerateImportanceGames.setEnabled(false);
        buttonLoadData.setEnabled(false);
        monteCarloButton.setEnabled(false);
        buttonStartingValues.setEnabled(false);
        buttonPlotFunction.setEnabled(false);
    }

    /**
     * Generate importance games for the estimation.
     */
    public void generateImportanceGames() {
        // generate importance games and calculate their equilibria
        append("Generating importance games...");
        labelStatus.setText("Status: Generating importance games...");
        SwingWorker worker = new SwingWorker() {

            public Object doInBackground() {
                importanceGamesList = new ArrayList<Game>();
                setBarMaximum(GameTheoryConstants.NUMBER_IMPORTANCE_GAMES);
                ArrayList<Future<Game>> workerList = new ArrayList<Future<Game>>();

                try {
                    for (int i = 0; i < GameTheoryConstants.NUMBER_IMPORTANCE_GAMES; i++) {
                        workerList.add(tpes.submit(new GenerateImportanceGameTask((gameList.get((int) Math.floor(gameList.size() * Math.random()))).deepCopy(), i)));
                    }
                    for (int i = 0; i < GameTheoryConstants.NUMBER_IMPORTANCE_GAMES; i++) {
                        Game g2 = workerList.get(i).get();
                        ArrayList equilibriaList = g2.getEquilibria();
                        if (equilibriaList.size() > 0 && Math.floor(equilibriaList.size() / 2.0) != (equilibriaList.size() / 2.0)) {
                            importanceGamesList.add(g2);
                            setBarValue(i + 1);
                        } else {
                            // System.out.println("Rejected: "+equilibriaList.size());
                            i--;
                        }
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }


                for (int i = 0; i < gameList.size(); i++) {
                    (gameList.get(i)).setImportanceGamesList(importanceGamesList);
                }

                try {
                    ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream("importanceGames.dat"));
                    out.writeObject(importanceGamesList.size());
                    for (int i = 0; i < importanceGamesList.size(); i++) {
                        out.writeObject(importanceGamesList.get(i));
                    }
                    out.close();
                } catch (Exception e) {
                    e.printStackTrace();
                }

                return null;
            }

            @Override
            public void done() {
                buttonExecuteMinimizationActionPerformed(null);
                // setButtonsReady();
                // append("OK\n");
                // labelStatus.setText("Status: Idle");

            }
        };
        setButtonsBusy();
        worker.execute();
    }

    /**
     * Entry point to the empirical application.
     *
     * @param args the command line arguments
     */
    public static void main(String args[]) {
        GenerateRecursiveList lister = new GenerateRecursiveList(GameTheoryConstants.NUM_PLAYERS, 2);
        GameTheoryConstants.actionList = lister.getCombinations();

        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
            e.printStackTrace();
        }

        final GameTheoryGUI go = new GameTheoryGUI();
        java.awt.EventQueue.invokeLater(new Runnable() {

            public void run() {
                go.setVisible(true);
            }
        });
    }

    /**
     * Set the maximum in the progress bar.
     *
     * @param i Highest value that the bar can take.
     */
    public void setBarMaximum(final int i) {
        Runnable doSetProgressBarValue = new Runnable() {

            public void run() {
                bar.setMaximum(i);
            }
        };
        SwingUtilities.invokeLater(doSetProgressBarValue);
    }

    /**
     * Sets progress bar value.
     *
     * @param i Value to set progress bar to.
     */
    public void setBarValue(final int i) {
        Runnable doSetProgressBarValue = new Runnable() {

            public void run() {
                bar.setValue(i);
            }
        };
        SwingUtilities.invokeLater(doSetProgressBarValue);
    }

    /**
     * Append a string to the primary text area.
     *
     * @param s String to be appended.
     */
    public void append(String s) {
        jt.append(s);
        jt.setCaretPosition(jt.getDocument().getLength());
    }

    // Variables declaration - do not modify//GEN-BEGIN:variables
    public javax.swing.JPanel BarPanel;
    public javax.swing.JPanel TextPanel;
    public javax.swing.JProgressBar bar;
    public javax.swing.JButton buttonExecuteMinimization;
    public javax.swing.JButton buttonGenerateImportanceGames;
    public javax.swing.ButtonGroup buttonGroupMinimizerType;
    public javax.swing.ButtonGroup buttonGroupTableView;
    public javax.swing.JButton buttonLoadData;
    public javax.swing.JButton buttonPlotFunction;
    public javax.swing.JButton buttonStartingValues;
    public javax.swing.JCheckBox checkBoxPlot;
    public javax.swing.JTable equilibriaTable;
    public javax.swing.JTextField fieldEfficientShifter;
    public javax.swing.JTextField fieldProfitScaleFactor;
    public javax.swing.JPanel graphPanel;
    public javax.swing.JTextField importanceBidCostField;
    public javax.swing.JLabel jLabel1;
    public javax.swing.JLabel jLabel10;
    public javax.swing.JLabel jLabel11;
    public javax.swing.JLabel jLabel12;
    public javax.swing.JLabel jLabel13;
    public javax.swing.JLabel jLabel14;
    public javax.swing.JLabel jLabel15;
    public javax.swing.JLabel jLabel17;
    public javax.swing.JLabel jLabel2;
    public javax.swing.JLabel jLabel3;
    public javax.swing.JLabel jLabel4;
    public javax.swing.JLabel jLabel5;
    public javax.swing.JLabel jLabel6;
    public javax.swing.JLabel jLabel7;
    public javax.swing.JLabel jLabel8;
    public javax.swing.JLabel jLabel9;
    public javax.swing.JPanel jPanel1;
    public javax.swing.JPanel jPanel2;
    public javax.swing.JRadioButton jRadioButton1;
    public javax.swing.JRadioButton jRadioButton2;
    public javax.swing.JScrollPane jScrollPane1;
    public javax.swing.JScrollPane jScrollPane2;
    public javax.swing.JScrollPane jScrollPane3;
    public javax.swing.JScrollPane jScrollPane4;
    public javax.swing.JSpinner jSpinner1;
    public javax.swing.JTable jTable1;
    public javax.swing.JTextArea jt;
    public javax.swing.JLabel labelStatus;
    public javax.swing.JButton monteCarloButton;
    public javax.swing.JPanel panelControl;
    public javax.swing.JPanel panelEquilibria;
    public javax.swing.JPanel panelEstimates;
    public javax.swing.JPanel panelPayoffs;
    public javax.swing.JRadioButton radioButtonMonteCarloData;
    public javax.swing.JSpinner spinnerMonteCarloNumObs;
    public javax.swing.JSpinner spinnerMonteCarloNumberIterations;
    public javax.swing.JSpinner spinnerNumPlayers;
    public javax.swing.JSpinner spinnerNumberImportanceGames;
    public javax.swing.JSpinner spinnerNumberThreads;
    public javax.swing.JSpinner spinnerOutcomesPerGame;
    public javax.swing.JSpinner spinnerParameterPlot;
    public javax.swing.JTabbedPane tabbedPaneTable;
    public javax.swing.JTable tableResults;
    public javax.swing.JTextField textFieldPureShifter;
    // End of variables declaration//GEN-END:variables
}
