/*
 * gibbsLTE.java
 *
 * Created on May 10, 2004, 2:39 PM
 */
package mcmcSolver;

import JSci.maths.statistics.*;
import java.io.*;
import java.util.ArrayList;

/**
 * Implementation of the LTE optimization/estimation method proposed in Chernozhukov and Hong (2003).
 * This approach is very useful for optimizing badly-behaved objective functions, and for simultaneously
 * computing the solution and posterior distributions in well-behaved problems.  The regularity condition
 * is that the objective function have a limited chi-squared distribution, such as a log-likelihood or a properly
 * weighted GMM objective function.
 * <p>
 * There are three steps in the algorithm:
 * <ol>
 * <li> Pre-burn
 * <li> Variance tuning
 * <li> Final burn
 * </ol>
 * <p>
 * The pre-burn takes a default sampling variance and executes the algorithm.  This typically results in a run toward
 * the (local) minimum.  Once the minimum is found, the algorithm switches to a variance tuning stage, where the
 * acceptance/rejection ratio is set to somewhere around 0.234 by adjusting the variance of the proposal density.
 * Finally, the actual values are generated in the last stage.
 *
 * There are many options which are largely self-explanatory when cross-referenced against the code.  This version ("Quiet")
 * does not contain any mechanisms for reporting results to a GUI or the command line.
 *
 * @author  Stephen P. Ryan
 */
public class gibbsLTEGeneralizedQuiet {

    /**
     * Boolean indicating whether the algorithm should tune the variance of the proposal density.
     */
    static boolean TUNE_VARIANCE = false;
    /**
     * An option useful in cases where the output will be used as an input to set-identified inference methods, such as
     * Chernozhukov, Hong, and Tamer (2006).
     */
    static boolean GENERATE_SET_SAMPLES = false;
    /**
     * Switch to override the stochastic acceptance algorithm to only accept improvements in the objective function.
     */
    static boolean DOWNHILL_ONLY = false;  // this is if you want to only accept improvements
    /**
     * Switch to draw all parameters from the proposal density jointly instead of one at a time.  Generally much slower to converge to local
     * mimima than the marginal method.
     */
    static boolean DRAW_ALL_SIMULTANEOUSLY = false;
    /**
     * How many draws per parameter until the adaptive algorithm stops.
     */
    static int STOPPING_CRITERION = 200;
    /**
     * Switch to use an adaptive algorithm that runs until STOPPING_CRITERION number of parameters have been sampled without
     * an improvement in the objective function.
     */
    static boolean USE_ADAPTIVE_STOPPING_RULE = false;
    /**
     * Jama.Matrix containing the results of the MCMC chain.
     */
    Jama.Matrix chain;
    /**
     * Target acceptance/rejection level for proposal density tuning.
     */
    double ACCEPTANCE_RATIO = 0.234;
    /**
     * Minimum level of variance in proposal density.
     */
    double MINIMUM_VARIANCE = 0.1;
    /**
     * Vector of proposal density variances.
     */
    double[] variance;
    /**
     * Utility class for formatting numbers.
     */
    java.text.NumberFormat nf = java.text.NumberFormat.getInstance();
    private double[] lowestPoint;
    private double[] point;
    double lowestValue;
    static int NUM_INSTANCES = 0;

    /**
     * Constructor for LTE class.
     *
     * @param function An object which implements the <i>mcmcFunction</i> interface as the source of the objective function.
     * @param draws Number of final draws to take in the MCMC chain.
     * @param burnPeriod Number of initial draws in the burn-in period.
     * @param guess Initial guess for the unknown parameters.
     */
    public gibbsLTEGeneralizedQuiet(mcmcFunction function, int draws, int burnPeriod, double[] guess) {
        execute(function, draws, burnPeriod, guess);
    }

    /**
     * Primary method for constructing the LTE Markov chain.
     *
     * @param function An object implementing the <i>mcmcFunction</i> interface as the source of the objective function.
     * @param draws Number of final draws to take in the MCMC chain.
     * @param burnPeriod Number of initial draws in the burn-in period.
     * @param guess Initial guess for the unknown parameters.
     */
    private void execute(final mcmcFunction function, int draws, int burnPeriodPassed, double[] guess) {
        if (TUNE_VARIANCE && DOWNHILL_ONLY) {
            System.out.println("Mutually incompatible parameters: Tuning Variance and Downhill Only");
            System.exit(0);
        }

        final int burnPeriod = burnPeriodPassed;
        lowestPoint = new double[guess.length];
        updateLowestPoint(guess);
        lowestValue = function.objectiveFunction(guess);

        final int numParams = guess.length;
        point = new double[numParams];
        variance = new double[numParams];

        double[] count = new double[numParams];

        for (int i = 0; i < numParams; i++) {
            point[i] = guess[i];
            variance[i] = Math.max(MINIMUM_VARIANCE, Math.abs(point[i] / 30.0));
            count[i] = 0;
        }

        double fold = function.objectiveFunction(point);

        int ijk = 0;
        fold = function.objectiveFunction(point);
        ProbabilityDistribution normal = new NormalDistribution(0, 1);

        int t = 0;
        int drawsSinceImprovement = 0;
        int m = 0;

        /**
         * Initial burn-in period.
         */
        while (ijk < burnPeriod || (USE_ADAPTIVE_STOPPING_RULE && drawsSinceImprovement <= STOPPING_CRITERION)) {
            double[] pointOld = new double[numParams];
            t++;
            drawsSinceImprovement++;
            // pmUtility.prettyPrint(new Jama.Matrix(point, 1));
            for (int j = 1; j < numParams; j++) {
                double pt = point[j] + variance[j] * normal.inverse(Math.random());
                if (DRAW_ALL_SIMULTANEOUSLY) {
                    for (int lm = 1; lm < numParams; lm++) {
                        pointOld[lm] = point[lm];
                        point[lm] = point[lm] + variance[lm] * normal.inverse(Math.random());
                    }
                } else {
                    pointOld[j] = point[j];
                    point[j] = pt;
                }

                double fnew = function.objectiveFunction(point);
                // System.out.println("j: " + j + " old: " + (-fold) + " new: " + (-fnew) + " - " + pmUtility.stringPrettyPrint(new Jama.Matrix(point, 1)) + " ");
                double diff = fnew - fold;
                double draw = Math.log(Math.random());
                if (DOWNHILL_ONLY) {
                    draw = 0;
                }
                if (fnew > lowestValue) {
                    lowestValue = fnew;
                    updateLowestPoint(point);
                    // System.out.println(t + ",x[" + j + "]=" + pt + "," + fnew);
                    drawsSinceImprovement = 0;
                }
                if (draw < diff) {
                    fold = fnew;
                } else {
                    point[j] = pointOld[j];
                    if (DRAW_ALL_SIMULTANEOUSLY) {
                        for (int lm = 1; lm < numParams; lm++) {
                            point[lm] = pointOld[lm];
                        }
                    }
                }
                m++;
            }
            ijk++;
        }

        // select variance using the replace one at a time method
        boolean go = true;
        for (int i = 0; i < numParams; i++) {
            variance[i] = Math.max(MINIMUM_VARIANCE / 10.0, Math.abs(point[i] / 30.0));
        }

        /**
         * Variance tuning stage.
         */
        while (go && TUNE_VARIANCE) {
            for (int j = 1; j < numParams; j++) {
                double acceptInside = 0;
                double rejectInside = 0;
                for (int i = 0; i < 250; i++) {
                    double pt = point[j] + variance[j] * normal.inverse(Math.random());
                    double po = point[j];
                    point[j] = pt;
                    double fnew = function.objectiveFunction(point);
                    double diff = fnew - fold;
                    double draw = Math.log(Math.random());
                    if (DOWNHILL_ONLY) {
                        draw = 0;
                    }
                    if (fnew > lowestValue) {
                        lowestValue = fnew;
                        updateLowestPoint(point);
                    }
                    if (draw < diff) {
                        acceptInside++;
                        fold = fnew;
                    } else {
                        rejectInside++;
                        point[j] = po;
                    }
                    m++;
                }
                double acceptRate = acceptInside / (acceptInside + rejectInside);
                double buffer = 0.20;
                if (acceptRate > (1.0 + buffer) * ACCEPTANCE_RATIO) {
                    variance[j] *= 1.15;
                    count[j] = Math.max(0, count[j] - 1);
                }
                if (acceptRate < (1.0 - buffer) * ACCEPTANCE_RATIO) {
                    variance[j] *= 0.85;
                    count[j] = Math.max(0, count[j] - 1);
                }
                if (acceptRate == 0) {
                    variance[j] *= 0.5;
                }
                if (acceptRate == 1) {
                    variance[j] *= 2;
                }
                if (acceptRate <= ACCEPTANCE_RATIO * (1.0 + buffer) && acceptRate >= ACCEPTANCE_RATIO * (1.0 - buffer)) {
                    count[j]++;
                }
            }
            go = false;
            for (int i = 1; i < numParams; i++) {
                if (!go) {
                    if (count[i] >= 3) {
                        go = false;
                    } else {
                        go = true;
                    }
                }
            }
        }

        /**
         * Final run for chain generation.
         */
        chain = new Jama.Matrix(draws, numParams);
        int accept = 0;
        int reject = 0;

        int drawsSinceLastImprovement = 0;
        int i = -1;

        boolean continueFinalRun = true;

        while (continueFinalRun) {
            if (continueFinalRun) {
                i++;
                if (USE_ADAPTIVE_STOPPING_RULE && drawsSinceLastImprovement > STOPPING_CRITERION) {
                    continueFinalRun = false;
                }
                if (!USE_ADAPTIVE_STOPPING_RULE && i >= draws - 1) {
                    continueFinalRun = false;
                }
                t++;
                drawsSinceLastImprovement++;
                for (int j = 1; j < numParams; j++) {
                    // double pt = point[j]+cauchy.inverse(Math.random());

                    double pt = point[j] + variance[j] * normal.inverse(Math.random());
                    double po = point[j];
                    point[j] = pt;
                    double fnew = function.objectiveFunction(point);
                    double diff = fnew - fold;
                    double draw = Math.log(Math.random());
                    if (DOWNHILL_ONLY) {
                        draw = 0;
                    }
                    if (fnew > lowestValue) {
                        lowestValue = fnew;
                        updateLowestPoint(point);
                        drawsSinceLastImprovement = 0;
                    }
                    if (draw < diff) {
                        accept++;
                        fold = fnew;
                    } else {
                        reject++;
                        point[j] = po;
                    }
                    m++;
                }

                if (!USE_ADAPTIVE_STOPPING_RULE) {
                    for (int k = 0; k < numParams; k++) {
                        chain.set(i, k, point[k]);
                    }
                }
            }
        }
    }

    /**
     * Keeps track of the lowest point found so far.
     * @param x Lowest point to be stored.
     */
    private void updateLowestPoint(double[] x) {
        for (int i = 0; i < x.length; i++) {
            lowestPoint[i] = x[i];
        }
    }

    /**
     * Returns the MCMC chain generated by the LTE procedure.
     * @return LTE chain.
     */
    public Jama.Matrix getChain() {
        return chain;
    }

    /**
     * Returns the lowest point found during the LTE procedure.
     * @return Lowest point found in the LTE procedure.
     */
    public double[] getLowestPoint() {
        return lowestPoint;
    }
}
