/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package montecarlo;

import JSci.maths.statistics.NormalDistribution;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Random;

/**
 * Class containing logic for a two-player, two-action discrete game of complete information.  Data fields contain information on
 * utilities, payoffs, and equilibria associated with the game. Contains methods to generate utilities, calculate
 * equilibria, calculate probabilities and moments, and simulate outcomes.
 *
 * This class represents a game of entry with two players.  Outcomes are indexed from 0 to 3, corresponding to both enter, only player 1 enters, only player 2 enters, and both enter
 * respectively.
 *
 * @author Stephen P. Ryan
 */
public class Game implements Serializable {

    /**
     * List of equilibria associated with this game.
     */
    private ArrayList<Equilibrium> equilibriaList = new ArrayList<Equilibrium>();
    /**
     * Array containing utilities, indexed by action and player, respectively.
     */
    private double[][] utilities = new double[4][2];
    /**
     * Array containing payoffs, indexed by action and player, respectively.
     */
    private double[][] payoffs = new double[4][2];
    /**
     * Array containing equilibria probabilites, indexed by equilibrium and player respectively.
     */
    private double[][] equilibria;
    /**
     * Shifter of player 1's payoffs for one of the actions.
     */
    private double x1;
    /**
     * Shifter of player 2's payoffs for one of the actions.
     */
    private double x2;
    /**
     * Simulated outcome (data).
     */
    private int outcome;
    /**
     * Vector of true parameters.
     */
    private double[] truth;
    /**
     * Standard normal distribution used for several calcuations.
     */
    static NormalDistribution normal = new NormalDistribution(0, 1);
    /**
     * Probability in the denominator of the importance sampling weights.  This is calculated by evaluating the probability of the drawn game at the true parameters.
     */
    private double probabilityGameTruth;
    /**
     * Count of the number of equilibria associated with the game.
     */
    private int numEqu;
    /**
     * Random number generator for producing simulated outcomes.  This is seeded when the constructor is called.
     */
    private Random rng;

    /**
     * Primary constructor for a 2-by-2 game.
     * @param x Vector of primitives associated with the game.
     * @param randomSeed Seed for the random-number generator.
     */
    public Game(double[] x, long randomSeed) {
        /**
         * Seed the random number generator.
         */
        rng = new Random(randomSeed);

        /**
         * Generate observables.
         */
        generateCovariates();

        /**
         * Generate deterministic payoffs.
         */
        payoffs = generatePayoffs(x);

        /**
         * Set utilities, which are payoffs plus errors.
         */
        utilities = generateUtilities();

        /**
         * Calculate equilibria.
         */
        calculateEquilibria();

        /**
         * Simulate an outcome.
         */
        simulateOutcome(x);

        truth = new double[x.length];
        for (int i = 0; i < x.length; i++) {
            truth[i] = x[i];
        }
        probabilityGameTruth = probabilityGame(truth);
    }

    /**
     * Generate the deterministic payoffs for the game.
     *
     * @param parameters Shifters for deterministic payoffs for each agent.
     * @return Array with deterministic payoffs, indexed by action and player.
     */
    private double[][] generatePayoffs(double[] parameters) {
        double[][] payoffs2 = new double[4][2];
        payoffs2[0][0] = parameters[1] * x1 + parameters[2];
        payoffs2[0][1] = parameters[1] * x2 + parameters[2];
        payoffs2[1][0] = parameters[1] * x1;
        payoffs2[1][1] = 0;
        payoffs2[2][0] = 0;
        payoffs2[2][1] = parameters[1] * x2;
        payoffs2[3][0] = 0;
        payoffs2[3][1] = 0;
        return payoffs2;
    }

    /**
     * Generate utilities for game.
     *
     * @return Array of utilities, indexed by action and player.
     */
    private double[][] generateUtilities() {
        double[][] utilities2 = new double[4][2];
        for (int i = 0; i < 4; i++) {
            for (int j = 0; j < 2; j++) {
                utilities2[i][j] = payoffs[i][j];
            }
        }
        utilities2[0][0] = payoffs[0][0] + normal.inverse(rng.nextDouble());
        utilities2[0][1] = payoffs[0][1] + normal.inverse(rng.nextDouble());
        utilities2[1][0] = payoffs[1][0] + normal.inverse(rng.nextDouble());
        utilities2[2][1] = payoffs[2][1] + normal.inverse(rng.nextDouble());
        return utilities2;
    }

    /**
     * Draw payoff shifters.
     */
    private void generateCovariates() {
        x1 = normal.inverse(rng.nextDouble()) + 1;
        x2 = normal.inverse(rng.nextDouble()) + 1;
    }

    /**
     * Return GMM moments, as defined by matching the empirical frequency of an outcome to the probability
     * predicted by the model.
     *
     * @param x Vector of unknown parameters.
     * @param importanceGamesList ArrayList of importance games used in importance sampler.
     * @return Matrix of moments.
     */
    public Jama.Matrix getMoments(double[] x, ArrayList<Game> importanceGamesList) {
        Jama.Matrix moments = new Jama.Matrix(3, 1);
        for (int i = 0; i < 3; i++) {
            double pr = 0;
            for (Game ig : importanceGamesList) {
                pr += ig.getProbabilityAction(x, i) * ig.getImportanceWeight(x);
            }
            pr /= importanceGamesList.size();
            int indicator = 0;
            if (i == outcome) {
                indicator = 1;
            }
            moments.set(i, 0, indicator - pr);
        }
        return moments;
    }

    /**
     * Method to update moment vector, starting with an extant moment vector and adding to it.
     *
     * @param moments Passed Jama.Matrix to add the current set of moments to.
     * @param x Vector of unknown parameters.
     * @param importanceGamesList ArrayList of importance games used in constructing the probabilities.
     */
    public void updateMoments(Jama.Matrix moments, double[] x, ArrayList<Game> importanceGamesList) {
        for (int i = 0; i < 3; i++) {
            double pr = 0;
            for (Game ig : importanceGamesList) {
                pr += ig.getProbabilityAction(x, i) * ig.getImportanceWeight(x);
            }
            pr /= importanceGamesList.size();
            int indicator = 0;
            if (i == outcome) {
                indicator = 1;
            }
            moments.set(i, 0, moments.get(i, 0) + indicator - pr);
        }
    }

    /**
     * Method to calculate probability of observing a given outcome.
     *
     * @param x Vector of unknown parameters.
     * @param outcomeIndex Index of observed action.
     * @return The probability of observing a given action.
     */
    private double getProbabilityAction(double[] x, int outcomeIndex) {
        double pr = 0;
        double[] probEqu = getProbabilityDistributionEquilibria(x);
        for (int i = 0; i < numEqu; i++) {
            double p = equilibria[i][0];
            double q = equilibria[i][1];
            if (outcomeIndex == 0) {
                // in, in
                pr += probEqu[i] * p * q;
            }
            if (outcomeIndex == 1) {
                pr += probEqu[i] * p * (1.0 - q);
            }
            if (outcomeIndex == 2) {
                pr += probEqu[i] * (1.0 - p) * q;
            }
        }
        return pr;
    }

    /**
     * Returns importance weight given vector of parameters.
     *
     * @param x Vector of unknown parameters.
     * @return Importance weight.
     */
    private double getImportanceWeight(double[] x) {
        return probabilityGame(x) / probabilityGameTruth;
    }

    /**
     * Returns the joint probability of observing a given game.
     *
     * @param x Vector of unknown parameters.
     * @return Probability of observing a joint draw of utilities, given the observable covariates.
     */
    private double probabilityGame(double[] x) {
        double prob = 0;
        double[][] payoffsDeterministic = generatePayoffs(x);
        prob += normal.probability(utilities[0][0] - payoffsDeterministic[0][0]);
        prob += normal.probability(utilities[0][1] - payoffsDeterministic[0][1]);
        prob += normal.probability(utilities[1][0] - payoffsDeterministic[1][0]);
        prob += normal.probability(utilities[2][1] - payoffsDeterministic[2][1]);
        return prob;
    }

    /**
     * Computes the equilibria to the current game.  Equilibria are stored in equilibriaList.
     */
    private void calculateEquilibria() {
        /**
         *  First, check for pure strategy equilibria, then check to see if there is a mixed strategy equilibrium
         */
        double inin1 = utilities[0][0];
        double inin2 = utilities[0][1];
        double inout1 = utilities[1][0];
        double inout2 = 0;
        double outin1 = 0;
        double outin2 = utilities[2][1];
        double outout1 = 0;
        double outout2 = 0;
        if (inin1 > outin1 && inin2 > inout2) {
            // pure (in, in)
            Equilibrium e = new Equilibrium(1, 1);
            equilibriaList.add(e);
        }
        if (inout1 > outout1 && inout2 > inin2) {
            // pure (in, out)
            Equilibrium e = new Equilibrium(1, 0);
            equilibriaList.add(e);
        }
        if (outin1 > inin1 && outin2 > outout2) {
            // pure (out, in)
            Equilibrium e = new Equilibrium(0, 1);
            equilibriaList.add(e);
        }
        if (outout1 > inout1 && outout2 > outin2) {
            // pure (out, out)
            Equilibrium e = new Equilibrium(0, 0);
            equilibriaList.add(e);
        }
        if (equilibriaList.size() == 0 || equilibriaList.size() == 2) {
            // must be a mixed strategy
            double q = inout1 / (inout1 - inin1);
            double p = outin2 / (outin2 - inin2);
            Equilibrium e = new Equilibrium(p, q);
            equilibriaList.add(e);
        }

        /**
         * Now take that list and convert into a double array for speed
         */
        equilibria = new double[equilibriaList.size()][2];
        for (int i = 0; i < equilibriaList.size(); i++) {
            Equilibrium e = equilibriaList.get(i);
            equilibria[i][0] = e.getP();
            equilibria[i][1] = e.getQ();
        }
        numEqu = equilibriaList.size();
        equilibriaList = null;
    }

    /**
     * Method to calculate the distribution over equilibria.
     *
     * @param x Vector of unknown parameters.
     * @return An array with the probability of each equilibrium encoded as a double.
     */
    private double[] getProbabilityDistributionEquilibria(double[] x) {
        double sum = 0;
        double[] prob = new double[numEqu];
        for (int i = 0; i < numEqu; i++) {
            double p = equilibria[i][0];
            double q = equilibria[i][1];
            if (p > 0 && p < 1 && q > 0 && q < 1) {
                prob[i] = Math.exp(x[3]);
            } else {
                prob[i] = 1;
            }
            sum += prob[i];
        }
        for (int i = 0; i < numEqu; i++) {
            prob[i] /= sum;
        }
        return prob;
    }

    /**
     * Simulates an outcome of the game.
     *
     * @param x Vector of unknown parameters.
     * @return An index specifying which outcome was played.
     */
    public int simulateOutcome(double[] x) {
        // need to simulate the outcome, accounting for # of equilibria and the equ selection mechanism
        // puts weight on mixed strategies in logit form
        double[] prob = getProbabilityDistributionEquilibria(x);
        double draw = rng.nextDouble();
        double sumProb = 0;
        int indexEquilibrium = 0;
        boolean go = true;
        while (go) {
            if (go) {
                sumProb += prob[indexEquilibrium];
                if (draw < sumProb) {
                    go = false;
                } else {
                    indexEquilibrium++;
                }
            }
        }

        double p = equilibria[indexEquilibrium][0];
        double q = equilibria[indexEquilibrium][1];

        double draw1 = rng.nextDouble();
        double draw2 = rng.nextDouble();
        boolean enter1 = false;
        boolean enter2 = false;

        if (draw1 < p) {
            enter1 = true;
        }
        if (draw2 < q) {
            enter2 = true;
        }

        if (enter1 && enter2) {
            outcome = 0;
        }
        if (enter1 && !enter2) {
            outcome = 1;
        }
        if (!enter1 && enter2) {
            outcome = 2;
        }
        if (!enter1 && !enter2) {
            outcome = 3;
        }
        return outcome;
    }

    /**
     * Returns index of simulated outcome.
     * @return The index of the simulated outcome.
     */
    public int getOutcome() {
        return outcome;
    }

    /**
     * Sets index of simulated outcome.
     * @param outcome The index of the outcome to set.
     */
    public void setOutcome(int outcome) {
        this.outcome = outcome;
    }
}
