/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package montecarlo;

import java.awt.BorderLayout;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.UIManager;
import org.gridgain.grid.Grid;
import org.gridgain.grid.GridException;
import org.gridgain.grid.GridFactory;
import org.gridgain.grid.GridNodeMetrics;
import utility.JTextAreaAutoscroll;
import utility.pmUtility;

/**
 * Program control for replicating the Monte Carlo experiments in Bajari, Hong, and Ryan (2009).  The
 * program makes use of a Gridgain computing grid if available; see <a href="http://www.gridgain.com">www.gridgain.com</a>.
 * <p>
 * The program works in a few simple steps:
 * <ol>
 * <li> Initialize the graphical user interface (GUI).
 * <li> Initialize an ExecutorService using either Gridgain or a fixed thread pool.
 * <li> Pass MonteCarloTask objects to the ExecutorService for evaluation.
 * <li> Collect and report the results.
 * <li> Repeat as desired for different numbers of observations, importance games, etc.
 * </ol>
 * <p>
 * The results are reported in a GUI, and consist of three sets of numbers (means, medians, standard deviations, mean bias, median bias,
 * mean absolute deviations, and median absolute deviations):
 * <ol>
 * <li> Results using the identity matrix as weights in the GMM objective function.
 * <li> Results using the optimal weighting matrix calculated using the results from the first step.
 * <li> Summary statistics for the standard deviations of each LTE chain used in each replication of the Monte Carlo.
 * </ol>
 * <p>
 * The primary numbers reported in the paper are drawn from the second set of results.  The first set is reported for completeness, as
 * they are used to construct the second set of results.  The third set of results is reported as a consistency check on the second set of
 * results.  For large enough MCMC chain sizes, the mean of the standard deviations of the LTE chains should be equal to the standard
 * deviations reported in the second set of results.  These were calculated for diagnostic purposes, and are not used in the paper,
 * but are kept here for completeness.
 *
 * We use the LTE method from Chernozhukov and Hong (2003) for optimizing each run in the Monte Carlo.  The appendix of their paper contains
 * an overview of the algorithm.  Newton-based methods are also included and may be used by changing the code in MonteCarloTask.
 * <p>
 * An overview of the concurrency architecture in Java can be found at <a href="http://java.sun.com/docs/books/tutorial/essential/concurrency/highlevel.html">java.sun.com</a>.
 *
 * @author Stephen P. Ryan
 */
public class Main {

    /**
     * JLabel for reporting the current running status of the program.
     */
    JLabel statusLabel;
    /**
     * Instance of the localized number formatting utility class.
     */
    static public NumberFormat nf = NumberFormat.getInstance();
    /**
     * The true values of the primitives.
     */
    static public double[] trueX = {0, 2, -10, 1};
    /**
     * Number of parameters.
     */
    static public int numParams = 3;
    /**
     * Number of Monte Carlo runs to perform.
     */
    static public int numRuns = 500;
    /**
     * Number of threads to use if not using a Gridgain grid.
     */
    static public int NUM_THREADS = 8;
    /**
     * Boolean indicating whether or not the program should attempt to utilize a Gridgain-based ExecutorService when submitting tasks
     */
    static public boolean useGrid = true;

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        // set the look and feel to the system default
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
            e.printStackTrace();
        }

        // initialize the GUI for the Monte Carlo
        JFrame f = new JFrame("Bajari Hong Ryan Monte Carlo");
        JTextArea jt = new JTextAreaAutoscroll("Program to Produce the Monte Carlo results in Bajari, Hong, and Ryan (2009)\n\n");
        JPanel statusPanel = new JPanel(new BorderLayout());
        JProgressBar bar = new JProgressBar(0, numRuns);
        JLabel statusLabel = new JLabel("Initializing");
        statusPanel.add(bar, BorderLayout.WEST);
        statusPanel.add(statusLabel, BorderLayout.EAST);
        bar.setStringPainted(true);
        f.getContentPane().add(statusPanel, BorderLayout.SOUTH);
        jt.setEditable(false);
        f.getContentPane().add(new JScrollPane(jt), BorderLayout.CENTER);
        f.setBounds(815, 250, 800, 400);
        f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        f.setVisible(true);

        try {
            // initialize the ExecutorService to use the Gridgain if selected, and a default fixed thread pool Executor otherwise
            ExecutorService tpesMain;
            if (useGrid) {
                try {
                    GridFactory.start();
                    Grid grid = GridFactory.getGrid();
                    tpesMain = grid.newGridExecutorService();
                } catch (GridException ex) {
                    Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
                    tpesMain = Executors.newFixedThreadPool(NUM_THREADS);
                }
            } else {
                tpesMain = Executors.newFixedThreadPool(NUM_THREADS);
            }

            // report number of nodes discovered if using gridgain
            if (useGrid) {
                jt.append("Number of nodes in grid: " + GridFactory.getGrid().getAllNodes().size() + "\n");
            }

            // set initial timer
            long t1 = System.currentTimeMillis();

            // loop over the number of games (observations) to evaluate
            for (int numGames = 25; numGames <= 100; numGames *= 2) {
                // set the number of MCMC draws used in the LTE solution algorithm of Chernozhukov and Hong (2003)
                for (int numMCMCDraws = 5000; numMCMCDraws <= 5000; numMCMCDraws *= 2) {
                    jt.append("\nNum MCMC Draws: " + numMCMCDraws + "\n");
                    // set the number of importance games
                    for (int numImportanceGames = numGames; numImportanceGames <= numGames; numImportanceGames *= 2) {
                        // initialize containers for holding results
                        jt.append("****************** N = " + numGames + " / " + numImportanceGames + " ******************\n");
                        Jama.Matrix results = new Jama.Matrix(numRuns, numParams);     // store each mean of each parameter (using whichever weighting matrix I selected)
                        Jama.Matrix preResults = new Jama.Matrix(numRuns, numParams);  // store the result using the identity matrix
                        Jama.Matrix lteResults = new Jama.Matrix(numRuns, numParams);  // store the result using the optimal weighting matrix
                        Jama.Matrix lteMeans = new Jama.Matrix(numRuns, numParams);    // store each asymptotic variance of each parameter

                        // update status label
                        statusLabel.setText("n = " + numGames);

                        // initialize ArrayList for holding Future objects as they return from the ExecutorService
                        ArrayList<Future<ArrayList<Jama.Matrix>>> futureList = new ArrayList<Future<ArrayList<Jama.Matrix>>>();

                        // submit tasks to the executorservice, and capture the resulting Future objects in the ArrayList above
                        for (int r = 0; r < numRuns; r++) {
                            futureList.add(tpesMain.submit(new MonteCarloTask(numGames, numImportanceGames, numMCMCDraws)));
                        }

                        // retrieve results from the executorservice, and store in appropriate containers
                        int counter = 0;
                        try {
                            for (Future<ArrayList<Jama.Matrix>> fu : futureList) {
                                ArrayList<Jama.Matrix> p = fu.get();
                                Jama.Matrix preResultsRun = p.get(0);
                                Jama.Matrix resultsRun = p.get(1);
                                Jama.Matrix lteMeansRun = p.get(2);
                                Jama.Matrix lteResultsRun = p.get(3);
                                for (int i = 0; i < numParams; i++) {
                                    results.set(counter, i, resultsRun.get(0, i));
                                    preResults.set(counter, i, preResultsRun.get(0, i));
                                    lteResults.set(counter, i, lteResultsRun.get(0, i));
                                    lteMeans.set(counter, i, lteMeansRun.get(0, i));
                                }
                                counter++;
                                bar.setValue(counter);
                                bar.setString(counter + " / " + futureList.size());
                            }
                        } catch (Exception e) {
                            e.printStackTrace();
                        }

                        // output some summary stats
                        // mean, median, std. dev, mean bias, median bias, mse, mean ad, median ad
                        jt.append("Parameter\tMean\tMedian\tStdDev\tMean Bias\tMedian Bias\tMean AD\tMedian AD\n");
                        jt.append("Pre-Results Means:\n");
                        Main.outputResults(preResults, jt, numParams);
                        jt.append("Results Means:\n");
                        Main.outputResults(results, jt, numParams);
                        jt.append("LTE StdDevs:\n");
                        Main.outputResults(lteResults, jt, numParams);
                    }
                }
            }
            
            // report total runtime
            long t2 = System.currentTimeMillis();
            jt.append("Total Runtime: " + (t2 - t1) / 1000.0 + " seconds.\n");
            jt.setCaretPosition(jt.getDocument().getLength());

            // shut down the executorservice
            tpesMain.shutdown();
        } finally {
            // shut down the grid if used and report some grid metrics
            if (useGrid) {
                GridNodeMetrics metrics = GridFactory.getGrid().getLocalNode().getMetrics();
                jt.append(metrics + "\n");
                GridFactory.stop(true);
            }
        }
    }

    /**
     * A simple utility method for reporting summary statistics from a passed Jama.Matrix object.
     * @param x The matrix containing the data to be summarized.
     * @param jt The JTextArea where the resulting information should be sent.
     * @param numParams The number of parameters to be evaluated.
     */
    public static void outputResults(Jama.Matrix x, JTextArea jt, int numParams) {
        for (int i = 0; i < numParams; i++) {
            double mean = pmUtility.mean(x, i);
            double median = pmUtility.median(x, i);
            double stddev = pmUtility.standardDeviation(x, i);
            double meanBias = mean - trueX[i + 1];
            double medianBias = median - trueX[i + 1];
            double meanAD = pmUtility.meanAbsoluteDeviations(x, i, trueX[i + 1]);
            double medianAD = pmUtility.medianAbsoluteDeviations(x, i, trueX[i + 1]);
            jt.append("x[" + (i + 1) + "]:\t" + nf.format(mean) + "\t" + nf.format(median) + "\t" + nf.format(stddev) + "\t" + nf.format(meanBias) + "\t" + nf.format(medianBias) + "\t" + nf.format(meanAD) + "\t" + nf.format(medianAD) + "\n");
        }
    }
}
