/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package montecarlo;

import Jama.Matrix;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Random;
import java.util.concurrent.Callable;
import mcmcSolver.gibbsLTEGeneralizedQuiet;
import mcmcSolver.mcmcFunction;
import newtonSolver.Uncmin_f77;
import newtonSolver.Uncmin_methods;
import utility.pmUtility;

/**
 * Task for evaluating a run of the estimator on a single set of fake data.
 *
 * @author Stephen P. Ryan
 */
public class MonteCarloTask implements Callable<ArrayList<Matrix>>, Uncmin_methods, mcmcFunction, Serializable {

    /**
     * ArrayList containing the importance games for a given observation.
     */
    private ArrayList<Game> importanceGamesList;
    /**
     * ArrayList containing list of observations.
     */
    private ArrayList<Game> data;
    /**
     * ArrayList containing a list of importance games associated with each observation.
     */
    private ArrayList<ArrayList<Game>> importanceMasterList = new ArrayList<ArrayList<Game>>();
    /**
     * Weighting matrix in GMM.
     */
    Jama.Matrix W = new Jama.Matrix(3, 3);  // optimal weighting matrix for GMM
    /**
     * Switch indicating whether or not a Newton-based optimizer is used in the minimization.
     */
    boolean USE_NEWTON_OPTIMIZER = false;
    /**
     * Switch indicating whether a first-stage estimate is produced using the identity weighting matrix
     */
    boolean ESTIMATE_FIRST_STAGE = true;
    /**
     * Vector of true parameters.
     */
    static public double[] TRUE_PARAMETERS = {0, 2, -10, 1};
    /**
     * Number of games (observations) to simulate in the fake data.
     */
    private int NUMBER_GAMES;
    /**
     * Number of importance games per observation.
     */
    private int NUMBER_IMPORTANCE_GAMES;
    /**
     * Number of draws in the LTE algorithm used to produce estimates of the parameters.
     */
    private int NUMBER_MCMC_DRAWS;

    /**
     * Constructor for Monte Carlo Task.
     * @param numGames Number of observations to use in this Monte Carlo experiment.
     * @param numImportanceGames Number of importance games to use in this Monte Carlo experiment.
     * @param numMCMCDraws Number of MCMC draws to use the LTE estimator of Chernozhukov and Hong (2003).
     */
    public MonteCarloTask(int numGames, int numImportanceGames, int numMCMCDraws) {
        this.NUMBER_GAMES = numGames;
        this.NUMBER_IMPORTANCE_GAMES = numImportanceGames;
        this.NUMBER_MCMC_DRAWS = numMCMCDraws;
    }

    /**
     * This method is called by an ExecutorService, and is where the computational logic is contained.
     * This method initializes the data set and importance games, and then, depending on the options selected
     * by the user, uses a combination of Newton-based methods and the LTE method of Chernozhukov and Hong (2003)
     * to optimize the resulting econometric problem.  The method contains an option for using a two-stage GMM
     * estimator, where the second stage uses an optimal weighting matrix computed from a first-stage GMM estimator
     * using an identity weighting matrix.
     *
     * @return Returns an ArrayList with the various Matrix outputs from the optimization.  See below for a description of the matrices.
     */
    public ArrayList<Matrix> call() throws Exception {
        ArrayList<Matrix> resultArray = new ArrayList<Matrix>();
        int numParams = 3;  // two utility parameters and one equilibrium selection parameter

        double[] guess = new double[numParams + 1];
        double[] xpls = new double[numParams + 1];
        double[] fpls = new double[2];
        double[] gpls = new double[numParams + 1];
        int[] itrmcd = new int[2];
        double[][] a = new double[numParams + 1][numParams + 1];
        double[] udiag = new double[numParams + 1];
        double[] typsiz = new double[numParams + 1];
        double[] fscale = {0, 1.0};
        int[] method = {0, 2}; // 1 line search 2 double dogleg 3 more-hebdon
        int[] iexp = {0, 1}; // 1 if expensive to evaluate
        int[] msg = {0, 0};
        int[] ndigit = {0, 5}; // number of good digits in minimand
        int[] itnlim = {0, 50};
        int[] iagflg = {0, 0};
        int[] iahflg = {0, 0};
        double[] dlt = {0, 1};
        double[] gradtl = {0, 1E-5};
        double[] stepmx = {0, 1E3};
        double[] steptl = {0, 1E-5};

        Uncmin_f77 minimizer = new Uncmin_f77(false);

        // we use the true values as initial values in the first stage
        // alternatives in other applications is to use a private information game to generate starting values
        guess[1] = TRUE_PARAMETERS[1];
        guess[2] = TRUE_PARAMETERS[2];
        guess[3] = TRUE_PARAMETERS[3];

        // initialize the data and set of importance games that will be used in the estimator
        initializeGames();

        // matrix used to store the best parameter found using whichever weighting matrix I selected
        Jama.Matrix results = new Jama.Matrix(1, numParams);

        // matrix used to store the result using the identity matrix
        Jama.Matrix identityMatrixResults = new Jama.Matrix(1, numParams);

        // matrix used to store the standard deviation of each parameter in the optimally-weighted GMM optimization
        Jama.Matrix optimalMatrixStandardDeviations = new Jama.Matrix(1, numParams);

        // matrix used to store the standard deviation of each parameter in the optimally-weighted GMM optimization
        Jama.Matrix optimalMatrixMeans = new Jama.Matrix(1, numParams);

        /**
         * Estimate with the identity weighting matrix.
         */
        if (ESTIMATE_FIRST_STAGE) {
            if (USE_NEWTON_OPTIMIZER) {
                minimizer.optif9_f77(numParams, guess, this, typsiz, fscale, method, iexp, msg, ndigit, itnlim, iagflg, iahflg, dlt, gradtl, stepmx, steptl, xpls, fpls, gpls, itrmcd, a, udiag);
            } else {
                gibbsLTEGeneralizedQuiet lte = new gibbsLTEGeneralizedQuiet(this, 250, 0, guess);
                guess = lte.getLowestPoint();
            }
            for (int i = 0; i < numParams; i++) {
                identityMatrixResults.set(0, i, guess[i + 1]);
            }
        }

        /**
         * If desired, estimate with the optimal weighting matrix calculated from the previous stage.
         */
        boolean useOptimalWeights = true;
        if (useOptimalWeights) {
            calculateOptimalWeightingMatrix(guess);

            gibbsLTEGeneralizedQuiet lte = new gibbsLTEGeneralizedQuiet(this, NUMBER_MCMC_DRAWS, 0, guess);
            guess = lte.getLowestPoint();
            Jama.Matrix chain = lte.getChain();
            for (int i = 0; i < numParams; i++) {
                optimalMatrixStandardDeviations.set(0, i, pmUtility.standardDeviation(chain, i + 1));
                optimalMatrixMeans.set(0, i, pmUtility.mean(chain, i + 1));
            }
        }
        for (int i = 0; i < numParams; i++) {
            results.set(0, i, guess[i + 1]);
        }
        resultArray.add(identityMatrixResults);
        resultArray.add(results);
        resultArray.add(optimalMatrixMeans);
        resultArray.add(optimalMatrixStandardDeviations);

        return resultArray;
    }

    /**
     * Initialize the importance games and data
     */
    private void initializeGames() {
        /**
         * Importance games first
         */
        Random rng = new Random();
        for (int j = 0; j < NUMBER_GAMES; j++) {
            importanceGamesList = new ArrayList<Game>();
            for (int i = 0; i < NUMBER_IMPORTANCE_GAMES; i++) {
                importanceGamesList.add(new Game(TRUE_PARAMETERS, rng.nextLong()));
            }
            importanceMasterList.add(importanceGamesList);
        }

        /**
         * Now data
         */
        Random rng2 = new Random();
        data = new ArrayList<Game>();
        double[] empiricalDist = new double[4];
        for (int i = 0; i < NUMBER_GAMES; i++) {
            Game g = new Game(TRUE_PARAMETERS, rng2.nextLong());
            data.add(g);
            empiricalDist[g.getOutcome()]++;
        }
        Jama.Matrix empDist = new Jama.Matrix(empiricalDist, 1);
        empDist.timesEquals(1.0 / NUMBER_GAMES);

        /**
         * Set initial weighting matrix to be equal to identity matrix.
         */
        W = Jama.Matrix.identity(3, 3);
    }

    /**
     * Objective function to minimize.
     * @param x Vector of unknown parameters.
     * @return Returns weighted GMM objective function value.
     */
    public double f_to_minimize(double[] x) {
        Jama.Matrix moments = new Jama.Matrix(3, 1);

        for (Game g : data) {
            moments.plusEquals(g.getMoments(x, importanceGamesList));
        }

        moments.timesEquals(1.0 / data.size());
        return (((moments.transpose()).times(W)).times(moments)).get(0, 0);
    }

    /**
     * Gradient, left empty since we just use numerical equivalents.
     * @param x Vector of unknown parameters
     * @param g Vector containing gradient
     */
    public void gradient(double[] x, double[] g) {
    }

    /**
     * Hessian, left empty since we just use numerical equivalents.
     * @param x Vector of unknown parameters
     * @param h Array containing Hessian
     */
    public void hessian(double[] x, double[][] h) {
    }

    /**
     * Calculates the optimal weighing matrix.
     *
     * @param x Vector of unknown parameters.
     */
    private void calculateOptimalWeightingMatrix(double[] x) {
        // W = E(gg'), where g is moment vector
        Jama.Matrix variance = new Jama.Matrix(3, 3, 0);
        for (Game game : data) {
            Jama.Matrix g = new Jama.Matrix(3, 1);
            game.updateMoments(g, x, importanceGamesList);
            variance.plusEquals(g.times(g.transpose()));
        }
        variance.timesEquals(1.0 / data.size());
        W = variance.inverse();
    }

    /**
     * Method passed to LTE minimizer.  Transforms the objective function above to chi-squared form.
     *
     * @param x Vector of unknown parameters.
     * @return Objective function normalized for use in the LTE procedure.
     */
    public double objectiveFunction(double[] x) {
        return -(data.size() / (2.0)) * f_to_minimize(x);
    }

    /**
     * Returns the probability of accepting a draw from the proposal density.  Always equal to one in this setting,
     * in general can be used to rule out parts of the parameter space.
     *
     * @param x Vector of proposed parameters.
     * @return Probability of acceptance.
     */
    public double pi(double[] x) {
        return 1;
    }
}
