/*
 * pmUtility.java
 *
 * Created on April 24, 2003, 4:58 PM
 */
package utility;

import Jama.Matrix;
import java.util.*;

/**
 * A collection of static utility methods for dealing with Jama.Matrix objects.
 * 
 * @author  Stephen P. Ryan
 */
public class pmUtility {

    /**
     * Returns the mean absolute deviations of a column of a matrix against a number.
     *
     * @param x Matrix of data.
     * @param column Index of column to be used in calculation.
     * @param compareTo Value to compute absolute deviations against.
     * @return Mean absolute deviation.
     */
    public static double meanAbsoluteDeviations(Matrix x, int column, double compareTo) {
        Jama.Matrix d = new Jama.Matrix(x.getRowDimension(), 1);
        for (int i = 0; i < x.getRowDimension(); i++) {
            d.set(i, 0, Math.abs(x.get(i, column) - compareTo));
        }
        return mean(d, 0);
    }

    /**
     * Returns the median absolute deviation of a column of a matrix against a number.
     *
     * @param x Matrix of data.
     * @param column Index of column to be used in calculation.
     * @param compareTo Value to compute absolute deviation against.
     * @return Median absolute deviation.
     */
    public static double medianAbsoluteDeviations(Matrix x, int column, double compareTo) {
        Jama.Matrix d = new Jama.Matrix(x.getRowDimension(), 1);
        for (int i = 0; i < x.getRowDimension(); i++) {
            d.set(i, 0, Math.abs(x.get(i, column) - compareTo));
        }
        return median(d, 0);
    }

    /**
     * Returns the mean of a given column in a matrix.
     *
     * @param x Matrix of data.
     * @param columnToSort Index of column to be used in the calcuation.
     * @return Mean of indicated column from passed matrix.
     */
    public static double mean(Jama.Matrix x, int columnToSort) {
        double mean = 0;
        for (int i = 0; i < x.getRowDimension(); i++) {
            mean += x.get(i, columnToSort);
        }
        mean /= (double) x.getRowDimension();
        return mean;
    }

    /**
     * Returns the median of a given column in a matrix.
     *
     * @param x Matrix of data.
     * @param columnToSort Index of column to be used in the calculation.
     * @return Median of indicated column from passed matrix.
     */
    public static double median(Jama.Matrix x, int columnToSort) {
        ArrayList<Double> sorter = new ArrayList<Double>();
        double median = 0;
        for (int i = 0; i < x.getRowDimension(); i++) {
            sorter.add(new Double(x.get(i, columnToSort)));
        }
        Collections.sort(sorter);
        median = (sorter.get((int) Math.round(x.getRowDimension() / 2.0) - 1)).doubleValue();
        return median;
    }

    /**
     * Returns the standard deviation of a given column in a matrix.
     *
     * @param x Matrix of data.
     * @param col Index of column to be used in the calculation.
     * @return The standard deviation of a given column in the matrix.
     */
    public static double standardDeviation(Jama.Matrix x, int col) {
        double mean = 0;
        double result = 0;
        // find mean
        for (int i = 0; i < x.getRowDimension(); i++) {
            mean += x.get(i, col);
        }
        mean /= x.getRowDimension();
        // find deviation
        for (int i = 0; i < x.getRowDimension(); i++) {
            result += (x.get(i, col) - mean) * (x.get(i, col) - mean);
        }
        result /= (x.getRowDimension() - 1);
        return Math.sqrt(result);
    }

}
